"""Base class for KeyRate objects."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.typing import ArrayLike, NDArray

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.quantum import Detector


class KeyRate(metaclass=abc.ABCMeta):
    """Key rate base class."""

    def __init__(self, detector: Detector, **kwargs: Any) -> None:
        """Init of KeyRate.

        Args:
            detector: The detector used at Bob's side
            kwargs: Protocol specific input
        """
        self.detector = detector
        self.__dict__.update(**kwargs)

    @abc.abstractmethod
    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Find the parameter setting that achieves the highest possible key-rate.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value
            bounds: Bounds on search range, provided by list of tuples containing lower-
                and upper bounds.

        Returns:
            Optimized arguments and key-rate

        Raises:
            ValueError: when x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
        """

    @abc.abstractmethod
    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        """Extract compute rate parameters from array."""

    @abc.abstractmethod
    def compute_rate(
        self, mu: float | ArrayLike, attenuation: float, **kwargs: Any
    ) -> float:
        """Computes the key-rate given intensity-settings and an attenuation.

        Args:
            mu: Used intensities
            attenuation: Attenuation of the channel
            kwargs: Depending on protocol more arguments can be supplied

        Returns:
            key-rate

        Raises:
            ValueError: When mu is given with invalid dimensions
        """

    def _f(self, x: ArrayLike, args: dict[str, NDArray[np.float64]]) -> float:
        """This function is minimized."""
        args.update(**self._extract_parameters(x))  # type: ignore[arg-type]
        return -self.compute_rate(**args)  # type: ignore[arg-type]


class AsymptoticKeyRateEstimate(KeyRate, metaclass=abc.ABCMeta):
    """Asymptotic key rate base class."""

    @abc.abstractmethod
    def compute_rate(self, *, mu: float | ArrayLike, attenuation: float) -> float:  # type: ignore[override]
        """Computes the key-rate given intensity-settings and an attenuation.

        Args:
            mu: Used intensities
            attenuation: Attenuation of the channel

        Returns:
            key-rate

        Raises:
            ValueError: When mu is given with invalid dimensions
        """


class FiniteKeyRateEstimate(KeyRate, metaclass=abc.ABCMeta):
    """Finite key rate base class."""

    @abc.abstractmethod
    def compute_rate(  # type: ignore[override]
        self,
        *,
        mu: float | ArrayLike,
        attenuation: float,
        probability_basis_X: ArrayLike,
        probability_basis_Z: ArrayLike,
        n_X: int | None = None,
    ) -> float:
        """Computes the key-rate given intensity-settings and an attenuation.

        Args:
            mu: Used intensities
            attenuation: Attenuation of the channel
            probability_basis_X: Probabilities for each intensity in the X-basis
            probability_basis_Z: Probabilities for each intensity in the Z-basis
            n_X: Number of pulses in the X-basis. If not provided, will be estimated
                from gain, probability_basis_X and total number of pules.

        Returns:
            key-rate

        Raises:
            ValueError: When mu is given with invalid dimensions
        """



# type: ignore

"""Classes to perform key error rate estimate for the BB84 QKD protocol.

This code is based on TNO's BB84 key-rate paper (doi: `10.1007/s11128-021-03078-0`_):

.. _10.1007/s11128-021-03078-0: https://doi.org/10.1007/s11128-021-03078-0

Quantum Key Distribution (QKD) protocols rely on an entangled photon source that
produces entangled photon pairs, which are distributed over two parties. Both parties
randomly choose one of two predefined measurement bases. As the photons are entangled,
non-random results will only be obtained for specific combinations of basis choices.
Detection events are sifted, where only the detection events corresponding to events
where both parties measured in an appropriate basis should be kept. Part of the
resulting sifted key is made public which can be used to estimate the key error rate.

The famous BB84 protocol by Charles Bennett and Gilles Brassard for establishing a
secure key between two parties, usually called Alice and Bob. Alice prepares a quantum
state in one of four ways, and Bob measures the quantum state in one of two ways. Based
on the way of measuring alone, both Alice and Bob can establish a key (assuming
noiseless operations). Classical post-processing routines can correct potential errors
still occurring and can detect eavesdroppers.

We consider three cases:

- Fully Asymptotic Key Rate
    Both the number of pulses and the number of used decoy states is infinite. Because
    of the asymptotic number of pulses and decoy states, we can simplify the
    computations and instead work with a single intensity setting which we vary.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bb84 import (
    ...     BB84FullyAsymptoticKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> fully_asymptotic_key_rate = BB84FullyAsymptoticKeyRateEstimate(detector=detector)
    >>> x, rate = fully_asymptotic_key_rate.optimize_rate(attenuation=0.2)
    >>> print(f"{x['mu']=}, {rate=}")  # doctest: +SKIP
    x['mu']=array([0.89982647]), rate=0.3494369214460756

- Asymptotic Key Rate
    Only the number of pulses is asymptotic, the number of decoy states is finite and
    chosen by the user. We have to optimize the probabilities for the X- and Z-basis
    for each intensity setting. So with two additional decoy states, we have three
    intensity settings to optimize and six probabilities in total. We use linear
    programs (LPs) for this.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bb84 import (
    ...     BB84AsymptoticKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> asymptotic_key_rate = BB84AsymptoticKeyRateEstimate(
    ...     detector=detector, number_of_decoy=3
    ... )
    >>> x, rate = asymptotic_key_rate.optimize_rate(
    ...     attenuation=0.2,
    ...     x0=[0.5, 0.3, 0.2, 0.4],
    ...     bounds=[(0.1, 0.9), (0.1, 0.9), (0.1, 0.9), (0.1, 0.9)],
    ... )
    >>> print(f"{x['mu']=}, {rate=}")  # doctest: +SKIP
    x['mu']=array([0.8998254 , 0.45375467, 0.21464554, 0.13087223]), rate=0.3494368339817865

- Finite Key Rate:
    Both the number of pulses and the number of decoy states is finite and chosen by
    the user. The approach is similar to the asymptotic key rate module, however we
    have to take finite-key effects into account. We compute bounds on the effect of
    the finite key size and we use security parameters to impose a degree of certainty
    of these bounds.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bb84 import (
    ...     BB84FiniteKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> finite_key_rate = BB84FiniteKeyRateEstimate(detector=detector)
    >>> x, rate = finite_key_rate.optimize_rate(attenuation=0.2)
    >>> print(f"{x['mu']=}, {rate=}")  # doctest: +SKIP
    x['mu']=array([9.80057214e-01, 5.76948581e-02, 2.95649571e-06]), rate=0.3359095058745817
"""  # noqa: E501

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

import numpy as np
import scipy
import scipy.stats
from numpy.typing import ArrayLike, NDArray

from tno.quantum.communication.qkd_key_rate._utils import binary_entropy as h
from tno.quantum.communication.qkd_key_rate.quantum._config import (
    LP_CONFIG,
    NLP_CONFIG,
    OptimizationError,
)
from tno.quantum.communication.qkd_key_rate.quantum._keyrate import (
    AsymptoticKeyRateEstimate,
    FiniteKeyRateEstimate,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.quantum import Detector

warnings.filterwarnings(
    "ignore", message="delta_grad == 0.0. Check if the approximated function is linear."
)


def ensure_probability(p: ArrayLike) -> NDArray[np.float64]:
    """Ensure that we have a probability between zero and one.

    Other functions will otherwise throw an error.

    Args:
        p: Probability to be mapped to range $[0, 1]$.

    Returns:
        Probability
    """
    return np.clip(p, 0, 1)


def compute_gain_and_error_rate(
    detector: Detector, mu: ArrayLike, attenuation: float
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Computes the total gain and error rate of the channel.

    Args:
        detector: The used detector on Bob's side
        mu: Intensities of the laser
        attenuation: Loss of the channel

    Returns:
        - The gain per intensity. The probability of an event, given a pulse.
        - The error rate per intensity.
    """
    dark_count_rate = detector.dark_count_rate
    efficiency_bob = detector.efficiency_party
    polarization_drift = detector.polarization_drift

    mu = np.atleast_1d(mu)

    efficiency_channel = np.power(10, -attenuation / 10)
    coefficient = -1 * mu * efficiency_bob * efficiency_channel

    # Compute the overall gain and error rate.
    # For stability and accuracy reasons, the error_rate is computed in two steps
    gain = 1 - np.power(1 - dark_count_rate, 2) * np.exp(coefficient)
    error_rate_denominator = 2 * gain
    error_rate_numerator = (
        1
        + (1 - dark_count_rate)
        * (
            np.exp(coefficient * np.power(np.cos(polarization_drift), 2))
            - np.exp(coefficient * np.power(np.sin(polarization_drift), 2))
        )
        - np.exp(coefficient) * np.power(1 - dark_count_rate, 2)
    )
    error_rate = error_rate_numerator / error_rate_denominator
    error_rate[np.isnan(error_rate)] = 0
    return gain, error_rate


def lower_bound_matrix_gain(
    max_num_photons: int, mus: ArrayLike
) -> NDArray[np.float64]:
    """Computes a lower bound on the likeliness of the number of photons per intensity.

    Args:
        max_num_photons: Maximum on the number of photons per pulse to consider
        mus: All used intensities
    """
    mus = np.atleast_1d(mus)
    i, mu = np.meshgrid(range(max_num_photons + 1), mus)
    return scipy.stats.poisson.pmf(i, mu)


def bound_f(
    number_of_pulses: int, probability: float, epsilon: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Computes a bound used in the finite key-rate computations.

    Args:
        number_of_pulses: Number of pulses considered
        probability: Probability of the event for which the bound-term is computed
        epsilon: Security-parameter

    Returns:
        Bounds to be used in finite-key rate computations.
    """
    return -np.log(epsilon) * (
        1 + np.sqrt(1 - 2 * probability * number_of_pulses / np.log(epsilon))
    )


def delta(n_x: int, n_z: int, e1: float) -> float:
    """Computes bound based on the number of pulses sent and epsilon-security value.

    Args:
        n_x: Number of pulses in the X-basis
        n_z: Number of pulses in the Z-basis
        e1: Epsilon-security parameter

    Returns:
        Bound.
    """
    return float(
        np.sqrt(np.max([-(n_x + n_z) * (n_x + 1) * np.log(e1) / (2 * n_z * n_x**2), 0]))
    )


def delta_ec(p_abort: float, n_x: int) -> float:
    """Computes a bound on the losses due to error correction.

    Args:
        p_abort: Abort probability, used if there are too many errors
        n_x: Number of pulses in the X-basis

    Returns:
        Bound.
    """
    return float(np.sqrt(np.log(2 / p_abort) * 3 * np.log2(5) ** 2 / n_x))


def solve_lp(
    target_vacuum: float,
    target_single: float,
    mu: NDArray[np.float64],
    program_coefficients: NDArray[np.float64],
    max_num_photons: int,
) -> scipy.optimize.OptimizeResult:
    """Solves the linear program (LP) for the asymptotic case.

    Args:
        target_vacuum: Coefficient for the vacuum state term
        target_single: Coefficient for the single photon state terms
        mu: The used intensity
        program_coefficients: The coefficients in the LP, e.g., gain,
            error-rate or their product
        max_num_photons: The number of photons at which the sums are cut

    Raises:
        ValueError: If intensity not same size as program coefficients.
        OptimizationError: in case no solution was found for the LP.
    """
    if len(mu) != len(program_coefficients):
        error_msg = "Intensity not compatible with program coefficients."
        raise ValueError(error_msg)

    # Determine the LP-constraints
    Lb = lower_bound_matrix_gain(max_num_photons, mu)
    Uv = 1 - np.sum(Lb, axis=1)

    lower_bound = np.zeros(max_num_photons + 1)
    upper_bound = np.ones(max_num_photons + 1)

    args = dict(  # noqa: C408
        c=np.hstack((target_vacuum, target_single, np.zeros(max_num_photons - 1))),
        A_ub=np.vstack((Lb, -Lb)),
        b_ub=np.concatenate((program_coefficients, -program_coefficients + Uv)),
        bounds=np.vstack((lower_bound, upper_bound)).T,
    )

    # Optimize the LP
    result = scipy.optimize.linprog(**args, **LP_CONFIG)
    if result.status != 0:
        error_msg = f"No solution was found for the linear problem. {result.message}"
        raise OptimizationError(error_msg)
    return result


def solve_finite_lp(  # noqa: PLR0913
    target_vacuum: float,
    target_single: float,
    probabilities_intensity_j: NDArray[np.float64],
    mu: NDArray[np.float64],
    max_num_photons: int,
    number_of_pulses: int,
    observed_count: NDArray[np.float64],
    epsilon_mu_j: NDArray[np.float64],
    epsilon_num_photons_M: float,
    epsilon_num_photons_M_in_basis_B: NDArray[np.float64],
) -> scipy.optimize.OptimizeResult:
    r"""Solves the linear program (LP) for the finite case.

    Args:
        target_vacuum: Coefficient for the vacuum state term
        target_single: Coefficient for the single photon state terms
        probabilities_intensity_j: Probability for each decoy state
        mu: The used intensity
        max_num_photons: The number of photons at which the sums are cut
        number_of_pulses: Number of pulses sent in specific basis
        observed_count: Number of pulses observed in specific basis per intensity
        epsilon_mu_j: Epsilon terms for the intensities
        epsilon_num_photons_M: Epsilon terms for the number of photons
        epsilon_num_photons_M_in_basis_B: Epsilon terms for the number of photons in
            basis B

    Returns:
        Number of usable pulses

    The variables in the LP are
        - $n_0$: Number of vacuum pulses
        - $n_1$: Number of single photon pulses
        - $\ldots$
        - $n_M$: Number of pulses with M pulses
        - $\delta_{j,1}$:  Deviation for intensity 1
        - $\delta_{j,2}$:  Deviation for intensity 2
        - $\ldots$
        - $\delta_{j,m}$:  Deviation for intensity m
    """
    # Total number of events observed
    observed_count = np.asarray(observed_count)
    observed_total = observed_count.sum()

    # Compute the probability to have an $m$ photon state, given you use intensity $j$
    probability_m_photon_state_given_intensity_j = scipy.stats.poisson.pmf(
        *np.meshgrid(range(max_num_photons + 1), mu)
    )
    # And compute the probability to have an $m$ photon state at all
    # (independent on the used intensity)
    probabilities_intensity_j = np.asarray(probabilities_intensity_j)
    probability_m_photon_state = probabilities_intensity_j.dot(
        probability_m_photon_state_given_intensity_j
    )
    probability_m_photon_state = np.clip(probability_m_photon_state, 1e-150, 1)

    # Similarly, compute the other conditional probabilities:
    # The probability to have intensity $j$, given that you send an $m$ photon state
    probability_intensity_j_given_m_photon_state = (
        probability_m_photon_state_given_intensity_j
        * np.outer(probabilities_intensity_j, 1 / probability_m_photon_state)
    )

    number_of_intensities = len(mu)
    # Compute the probability that a photon state is send out with more than
    # the defined maximum number of photons.
    tail_probability_maximum_photons = float(
        scipy.stats.poisson.sf(max_num_photons + 1, mu).dot(probabilities_intensity_j)
    )

    # Bound on the contribution for the number of photons above the cut
    lambda_num_photons_cut = (
        tail_probability_maximum_photons * number_of_pulses
        + bound_f(
            number_of_pulses,
            tail_probability_maximum_photons,
            epsilon_num_photons_M_in_basis_B,
        )
    )

    # Below, constraints are defined. The first two letters of the variables
    # define the type of constraint: lb = lower bound, ub = upper bound, eq = equality
    # Constraints on the number of $m$ photon states per intensity $j$
    lb_num_photons_m_given_intensity_j_lhs = -np.hstack(
        (probability_intensity_j_given_m_photon_state, np.eye(number_of_intensities))
    )
    lb_num_photons_m_given_intensity_j_rhs = -observed_count
    ub_num_photons_m_given_intensity_j_lhs = np.hstack(
        (probability_intensity_j_given_m_photon_state, np.eye(number_of_intensities))
    )
    ub_num_photons_m_given_intensity_j_rhs = observed_count + lambda_num_photons_cut

    # Constraints on the number of $m$ photon states
    lb_num_photons_m_lhs = -np.eye(
        max_num_photons + 1, max_num_photons + number_of_intensities + 1
    )
    lb_num_photons_m_rhs = np.zeros(max_num_photons + 1)
    ub_num_photons_m_lhs = np.eye(
        max_num_photons + 1, max_num_photons + number_of_intensities + 1
    )
    ub_num_photons_m_rhs = np.minimum(
        (
            probability_m_photon_state * number_of_pulses
            + bound_f(
                number_of_pulses,
                probability_m_photon_state,
                epsilon_num_photons_M,
            )
            * number_of_pulses
        ),
        np.repeat(observed_total, max_num_photons + 1),
    )

    # Constraints on the states where intensity $j$ is used
    intensity_j_rhs = np.sqrt(-np.log(epsilon_mu_j / 2) * observed_total / 2)
    intensity_j_lhs = np.eye(
        number_of_intensities,
        max_num_photons + number_of_intensities + 1,
        max_num_photons + 1,
    )
    lb_intensity_j_rhs = intensity_j_rhs
    ub_intensity_j_rhs = intensity_j_rhs
    lb_intensity_j_lhs = intensity_j_lhs
    ub_intensity_j_lhs = -intensity_j_lhs

    # Constraint that the states per intensity should together sum to the observed count
    eq_sum_of_intensities_j_lhs = np.hstack(
        (np.zeros(max_num_photons + 1), np.ones(number_of_intensities))
    )
    eq_sum_of_intensities_j_rhs = 0

    args = dict(  # noqa: C408
        c=np.hstack(
            (
                target_vacuum,
                target_single,
                np.zeros(max_num_photons + number_of_intensities - 1),
            )
        ),
        A_ub=np.vstack(
            (
                lb_num_photons_m_given_intensity_j_lhs,
                ub_num_photons_m_given_intensity_j_lhs,
                lb_num_photons_m_lhs,
                ub_num_photons_m_lhs,
                lb_intensity_j_lhs,
                ub_intensity_j_lhs,
            )
        ),
        b_ub=np.hstack(
            (
                lb_num_photons_m_given_intensity_j_rhs,
                ub_num_photons_m_given_intensity_j_rhs,
                lb_num_photons_m_rhs,
                ub_num_photons_m_rhs,
                lb_intensity_j_rhs,
                ub_intensity_j_rhs,
            )
        )
        / number_of_pulses,
        A_eq=np.atleast_2d(eq_sum_of_intensities_j_lhs),
        b_eq=eq_sum_of_intensities_j_rhs / number_of_pulses,
        bounds=(None, None),
    )

    result = scipy.optimize.linprog(**args, **LP_CONFIG)
    if result.status != 0:
        error_msg = f"No solution was found for the linear problem. {result.message}"
        raise OptimizationError(error_msg)
    return result


class BB84FullyAsymptoticKeyRateEstimate(AsymptoticKeyRateEstimate):
    """The situation for an asymptotic number of intensity settings and pulses.

    In the fully asymptotic case we only have to consider a single intensity
    """

    def __init__(self, detector: Detector, **kwargs: Any) -> None:
        """Init of the BB84FullyAsymptoticKeyRateEstimate.

        Args:
            detector: The detector used at Bob's side
            kwargs: Protocol specific input
        """
        super().__init__(detector=detector, args=kwargs)

    def compute_rate(self, mu: float | ArrayLike, attenuation: float) -> float:  # type: ignore[override]
        """Computes the key-rate given an intensity and an attenuation.

        Only the vacuum states and single photon states can be safely used.
        The error rate for vacuum states is 0.5. For single photon states we
        must upper bound it.

        Args:
            mu: Intensity
            attenuation: Attenuation

        Returns:
            Key-rate
        """
        mu = np.atleast_1d(mu)
        detector = self.detector

        efficiency_bob = detector.efficiency_party
        efficiency_channel = np.power(10, -attenuation / 10)
        efficiency_system = efficiency_bob * efficiency_channel

        gain, error_rate = compute_gain_and_error_rate(detector, mu, attenuation)

        # Compute the relevant terms for vacuum and single photon states
        yield_vacuum = 1 - np.power(1 - detector.dark_count_rate, 2)
        yield_single = 1 - (
            np.power(1 - detector.dark_count_rate, 2) * (1 - efficiency_system)
        )

        yield_times_error_single = (
            yield_single
            - (1 - detector.dark_count_rate)
            * efficiency_system
            * np.cos(2 * detector.polarization_drift)
        ) / 2
        error_single = yield_times_error_single / yield_single
        # The probabilities of having precisely zero or one photon in a state
        (
            probability_vacuum_basis_X,
            probability_single_basis_X,
        ) = scipy.stats.poisson.pmf([0, 1], mu[0])

        # Pulses in Z-basis are used for error estimation, the pulses in the
        # X-basis are used to obtain key-material
        h_entropy_single_basis_Z = 1 - h(error_single)

        gain_times_error_basis_X = gain * h(error_rate)

        return float(
            probability_vacuum_basis_X * yield_vacuum
            + (probability_single_basis_X * yield_single * h_entropy_single_basis_Z)
            - gain_times_error_basis_X
        )

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        return {"mu": x}

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value, default value [0.5] is chosen.
            bounds: Bounds on search range, default [(0.1, 0.9)]

        Raises:
            ValueError: When x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.

        Returns:
            Optimized intensity and key-rate
        """
        if bounds is None:
            # The lower and upper bound on the laser intensity
            lower_bound = np.array([0.1])
            upper_bound = np.array([0.9])
        else:
            if len(bounds) != 1:
                error_msg = (
                    "Invalid dimensions input bounds. Expected 1 upper- and lower"
                    f" bound but received {len(bounds)} bounds."
                )
                raise ValueError(error_msg)
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
        bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        if x0 is None:
            x0 = [0.5]
        elif len(np.asarray(x0)) != 1:
            error_msg = (
                "Invalid number of inputs. Expected 1 intensity."
                f" Received {len(np.asarray(x0))} intensities."
            )
            raise ValueError(error_msg)

        args = {"attenuation": attenuation}
        res = scipy.optimize.minimize(
            self._f, x0, args=args, bounds=bounds, **NLP_CONFIG
        )
        mu = res.x
        rate = np.atleast_1d(-res.fun)[0]
        if rate < 0:
            error_msg = "Optimization resulted in a negative key rate."
            raise ValueError(error_msg)
        return self._extract_parameters(mu), float(rate)


class BB84AsymptoticKeyRateEstimate(AsymptoticKeyRateEstimate):
    """The situation for an asymptotic number of pulses.

    We consider a fixed number of intensities (number_of_decoy + 1)
    """

    def __init__(
        self, detector: Detector, number_of_decoy: int = 2, **kwargs: Any
    ) -> None:
        """Init of BB84AsymptoticKeyRateEstimate.

        Args:
            detector:  The detector used at Bob's side
            number_of_decoy: Number of decoy intensities used
            kwargs: protocol specific arguments.
        """
        super().__init__(detector=detector, args=kwargs)
        self.number_of_decoy = number_of_decoy
        self.last_positive: float = -1
        self.last_x: NDArray[np.float64]

    def _compute_last_positive_distance(self, x: NDArray[np.float64]) -> float:
        """Computes the last positive distance.

        The optimization routine sometimes considers a parameter setting
        outside of the valid region. This function is used to push the
        parameters back to the valid regime.
        """
        if self.last_positive > -1:
            return self.last_positive - float(np.linalg.norm(x - self.last_x))
        return self.last_positive

    def compute_rate(self, mu: float | ArrayLike, attenuation: float) -> float:  # type: ignore[override]
        """Computes the key-rate given intensity-settings and an attenuation.

        Args:
            mu: Intensity
            attenuation: Attenuation

        Returns:
            Key-rate
        """
        mu = np.atleast_1d(mu)
        x = np.asarray(mu)

        if x.min() < 0 or x.max() > 1:
            # If the variable x is outside the possible range, push it back
            return self._compute_last_positive_distance(x)

        # Maximum on the number of photons per pulse to consider
        max_num_photons = np.max((int(scipy.stats.poisson.isf(1e-12, mu.max())), 5))

        gain, error_rate = compute_gain_and_error_rate(self.detector, mu, attenuation)

        # Compute the yield and error rate for single photon pulses in the Z basis
        yield_single_basis_Z = ensure_probability(
            solve_lp(
                target_vacuum=0,
                target_single=1,
                mu=mu,
                program_coefficients=gain,
                max_num_photons=max_num_photons,
            )["fun"]
        )
        yield_times_error_basis_Z = ensure_probability(
            -solve_lp(
                target_vacuum=0,
                target_single=-1,
                mu=mu,
                program_coefficients=gain * error_rate,
                max_num_photons=max_num_photons,
            )["fun"]
        )

        if yield_single_basis_Z == 0:
            error_single_basis_Z = 1e-19
        else:
            error_single_basis_Z = np.clip(
                yield_times_error_basis_Z / yield_single_basis_Z, 1e-19, 0.5
            )
        h_entropy_single_basis_Z = 1 - h(error_single_basis_Z)

        # Get the probability for vacuum and single photon pulses in the X basis
        (
            probability_vacuum_basis_X,
            probability_single_basis_X,
        ) = scipy.stats.poisson.pmf([0, 1], mu[0])

        # Compute the yield for vacuum and single photon pulses in the X basis
        yield_vacuum_single_basis_X = float(
            solve_lp(
                target_vacuum=probability_vacuum_basis_X,
                target_single=probability_single_basis_X * h_entropy_single_basis_Z,
                mu=mu,
                program_coefficients=gain,
                max_num_photons=max_num_photons,
            )["fun"]
        )
        if np.isnan(yield_vacuum_single_basis_X) or np.isinf(
            yield_vacuum_single_basis_X
        ):
            return self._compute_last_positive_distance(x)

        # Determine the overall key-rate
        gain_times_error_basis_X = float(gain[0] * h(error_rate[0]))
        key_rate = yield_vacuum_single_basis_X - gain_times_error_basis_X

        self.last_positive = key_rate
        self.last_x = np.asarray(mu)

        return float(key_rate)

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        return {"mu": x}

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        For certain input parameters it can happen that the resulting lp problem is
        unfeasible. In that case the attenuation is slightly modified (``+1e-8``) in an
        attempt to obtain a feasible lp problem that can be solved.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value, default midpoint search bounds
            bounds: Bounds on search range

        Returns:
            Optimized intensity and key-rate

        Raises:
            ValueError: When x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
            OptimizationError: When lp solver is unsuccessful due to infeasible problem.
                Multiple attempts are made with slightly modified  attenuation before
                error is raised.
        """
        if bounds is None:
            # Lower and upper bounds on the considered laser intensities
            lower_bound = 0.00000001 * np.ones(self.number_of_decoy + 1)
            lower_bound[0] = 0.2  # Help the optimizer
            upper_bound = 0.95 * np.ones(self.number_of_decoy + 1)
            upper_bound[-1] = 0.3  # Help the optimizer
        else:
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
            if len(bounds) != self.number_of_decoy + 1:
                error_msg = (
                    f"Invalid dimensions input bounds. Expected "
                    f"{self.number_of_decoy + 1} upper- and lower bounds. Received "
                    f"{len(bounds)} bounds."
                )
                raise ValueError(error_msg)
        optimize_bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        less_one = 1 - 1e-16

        self.last_positive = -1
        # If no ansatz is given, choose the mean of the bounds.
        if x0 is None:
            x0 = (lower_bound + upper_bound) / 2
        x0 = np.asarray(x0)
        if len(x0) != self.number_of_decoy + 1:
            error_msg = (
                "Invalid number of inputs. Expected "
                f"{self.number_of_decoy + 1} intensities. Received "
                f"{len(x0)} intensities."
            )
            raise ValueError(error_msg)
        self.last_x = x0

        # The constraints enforce the intensities to be decreasing
        b = np.repeat(-np.inf, self.number_of_decoy)
        A = np.triu(
            np.ones((self.number_of_decoy, self.number_of_decoy + 1)), 1
        ) - less_one * np.eye(self.number_of_decoy, self.number_of_decoy + 1)
        B = np.zeros(self.number_of_decoy)
        constraint = scipy.optimize.LinearConstraint(A, b, B)

        args = {"attenuation": attenuation}
        num_attempts = 3
        for _ in range(num_attempts):  # Maximum 3 retries
            try:
                res = scipy.optimize.minimize(
                    self._f,
                    x0,
                    args=args,
                    constraints=(constraint),
                    bounds=optimize_bounds,
                    **NLP_CONFIG,
                )
            except OptimizationError:  # noqa: PERF203
                args["attenuation"] += 1e-8
            else:
                rate = np.atleast_1d(-res.fun)[0]
                if rate < 0:  # Retry when negative key-rate is found.
                    args["attenuation"] += 1e-8
                    continue
                return self._extract_parameters(res.x), float(rate)
        error_msg = "Unable to find solution for optimal key rate."
        raise OptimizationError(error_msg)


class BB84FiniteKeyRateEstimate(FiniteKeyRateEstimate):
    """BB84 protocol with a finite number of pulses.

    A fixed number of intensities is considered.
    The probabilities for both bases might vary.
    """

    def __init__(
        self,
        detector: Detector,
        number_of_pulses: int = int(1e12),
        number_of_decoy: int = 2,
        **kwargs: Any,
    ) -> None:
        """Init of the BB84FiniteKeyRateEstimate.

        Args:
            detector: The detector used at Bob's side
            number_of_pulses: Number of pulses sent
            number_of_decoy: Number of decoy intensities used
            kwargs: Any protocol dependent parameters.
        """
        super().__init__(detector=detector, args=kwargs)
        self.number_of_pulses = number_of_pulses
        self.number_of_decoy = number_of_decoy
        self.last_positive: float = -1
        self.last_x: NDArray[np.float64]

    def _compute_last_positive_distance(self, x: NDArray[np.float64]) -> float:
        """Computes the last positive distance.

        The optimization routine sometimes considers a parameter setting
        outside of the valid region. This function is used to push the
        parameters back to the valid regime.
        """
        if self.last_positive > -1:
            return self.last_positive - float(np.linalg.norm(x - self.last_x))
        return self.last_positive

    def compute_rate(  # type: ignore[override]
        self,
        mu: float | ArrayLike,
        attenuation: float,
        probability_basis_X: ArrayLike,
        probability_basis_Z: ArrayLike,
        n_X: int | None = None,
    ) -> float:
        """Compute the key-rate for a specific set of parameters.

        Args:
            mu: Used intensities
            attenuation: Attenuation of the channel
            probability_basis_X: Probabilities for each intensity in the X-basis
            probability_basis_Z: Probabilities for each intensity in the Z-basis
            n_X: Number of pulses in the X-basis.

        Returns:
            key-rate
        """
        # Security parameters
        probability_abort = 2e-50
        epsilon_security = 2e-50
        epsilon_correct = 2e-50
        epsilon_1 = 2e-55
        epsilon_2 = 2e-55
        epsilon_3 = 2e-55
        epsilon = 1e-7

        # The probabilities sum to 1, these are the conditional probabilities per basis
        probability_basis_X = np.asarray(probability_basis_X)
        probability_basis_Z = np.asarray(probability_basis_Z)
        probability_basis_X_normalized: NDArray[np.float64] = (
            probability_basis_X / probability_basis_X.sum()
        )
        probability_basis_Z_normalized: NDArray[np.float64] = (
            probability_basis_Z / probability_basis_Z.sum()
        )

        mu = np.asarray(mu)
        x = np.hstack((mu, probability_basis_X, probability_basis_Z))
        if x.min() < 0 or x.max() > 1:
            return self._compute_last_positive_distance(x)

        gain, error_rate = compute_gain_and_error_rate(self.detector, mu, attenuation)

        # Compute number of detection events per intensity per basis
        if n_X is None:
            n_X_observed_per_intensity: NDArray[np.float64] = (
                gain * ensure_probability(probability_basis_X) * self.number_of_pulses
            )
            n_Z_observed_per_intensity: NDArray[np.float64] = (
                gain * ensure_probability(probability_basis_Z) * self.number_of_pulses
            )
        else:
            n_X_observed_per_intensity = gain * probability_basis_X_normalized * n_X
            n_Z_observed_per_intensity = (
                gain * probability_basis_Z_normalized * (self.number_of_pulses - n_X)
            )

        n_X_observed = float(n_X_observed_per_intensity.sum())
        n_Z_observed = float(n_Z_observed_per_intensity.sum())

        if n_X_observed == 0 or n_Z_observed == 0:
            return self._compute_last_positive_distance(x)

        # And their corresponding number of errors
        number_of_errors_per_intensity_basis_X = error_rate * n_X_observed_per_intensity
        number_of_errors_per_intensity_basis_Z = error_rate * n_Z_observed_per_intensity
        number_of_errors_basis_X = float(number_of_errors_per_intensity_basis_X.sum())

        # Maximum on the number of photons per pulse to consider
        max_num_photons = int(
            np.max((int(scipy.stats.poisson.isf(1e-16, mu).max()), 5))
        )

        # Epsilon values to be considered in the optimization.
        # Values for the intensities and the number of photons per pulse are used,
        # as well as the number of photon per pulse given a basis
        epsilon_single_photon = 1e-7
        epsilon_intensity_j: NDArray[np.float64] = np.repeat(1e-8, mu.shape[0])
        epsilon_num_photons_M: NDArray[np.float64] = np.repeat(
            1e-8, int(max_num_photons + 1)
        )
        epsilon_num_photons_M_in_basis_B: NDArray[np.float64] = np.repeat(
            1e-8, int(mu.shape[0])
        )

        # The probability of a single photon pulse for each of the intensities
        probability_single_photon_per_intensity: NDArray[np.float64] = (
            scipy.stats.poisson.pmf(1, mu)
        )
        probability_single_photon_per_intensity: NDArray[np.float64] = (
            probability_single_photon_per_intensity
            / probability_single_photon_per_intensity.sum()
        )

        # Solve linear programs to get the number of single photon events
        # and the number of errors in these single photon events in the Z basis
        # For vacuum pulses, the error is 0.5
        (_, number_single_photon_events_in_basis_Z) = (
            solve_finite_lp(
                target_vacuum=0,
                target_single=1,
                probabilities_intensity_j=probability_basis_Z_normalized,
                mu=mu,
                max_num_photons=max_num_photons,
                number_of_pulses=n_Z_observed,
                observed_count=n_Z_observed_per_intensity,
                epsilon_mu_j=epsilon_intensity_j,
                epsilon_num_photons_M=epsilon_num_photons_M,
                epsilon_num_photons_M_in_basis_B=epsilon_num_photons_M_in_basis_B,
            )["x"][0:2]
            * n_Z_observed
        )
        (_, number_single_photon_errors_in_basis_Z) = (
            solve_finite_lp(
                target_vacuum=0,
                target_single=1,
                probabilities_intensity_j=probability_basis_Z_normalized,
                mu=mu,
                max_num_photons=max_num_photons,
                number_of_pulses=n_Z_observed,
                observed_count=number_of_errors_per_intensity_basis_Z,
                epsilon_mu_j=epsilon_intensity_j,
                epsilon_num_photons_M=epsilon_num_photons_M,
                epsilon_num_photons_M_in_basis_B=epsilon_num_photons_M_in_basis_B,
            )["x"][0:2]
            * n_Z_observed
        )
        if number_single_photon_events_in_basis_Z == 0:
            error_rate_single_photon_basis_Z = 1e-19
        else:
            error_rate_single_photon_basis_Z = np.clip(
                number_single_photon_errors_in_basis_Z
                / number_single_photon_events_in_basis_Z,
                1e-19,
                0.5,
            )

        # Solve the linear program for vacuum and single photon pulses in X-basis
        # Note the expression for target_single, this takes into account the
        # errors in the Z-basis and the number of pulses in both.

        res = solve_finite_lp(
            target_vacuum=1,
            target_single=1
            - h(
                error_rate_single_photon_basis_Z
                + delta(
                    n_X_observed,
                    n_Z_observed,
                    epsilon_single_photon,
                )
            ),
            probabilities_intensity_j=probability_basis_X_normalized,
            mu=mu,
            max_num_photons=max_num_photons,
            number_of_pulses=n_X_observed,
            observed_count=n_X_observed_per_intensity,
            epsilon_mu_j=epsilon_intensity_j,
            epsilon_num_photons_M=epsilon_num_photons_M,
            epsilon_num_photons_M_in_basis_B=epsilon_num_photons_M_in_basis_B,
        )

        (
            number_vacuum_events_basis_X,
            number_single_events_basis_X,
        ) = res["x"][0:2] * n_X_observed

        if np.isnan(res["fun"]) or np.isinf(res["fun"]):
            return self._compute_last_positive_distance(x)

        # With the results from the LP, we can determine the number of usable pulses
        number_vacuum_single_pulses = (
            number_vacuum_events_basis_X
            + number_single_events_basis_X
            - number_single_events_basis_X
            * (
                h(
                    number_single_photon_errors_in_basis_Z
                    + delta(
                        n_X_observed,
                        n_Z_observed,
                        epsilon_1,
                    )
                )
            )
        )

        error_rate_basis_X = number_of_errors_basis_X / n_X_observed
        usable_pulses_lp = (
            number_vacuum_single_pulses
            - n_X_observed
            * (h(error_rate_basis_X) + delta_ec(probability_abort, n_X_observed))
            + np.log2(
                epsilon_correct
                * np.power(epsilon_2 * epsilon_3 * (epsilon_security - epsilon), 2)
            )
            - 1
        )

        # With which we can determine the key-rate
        key_rate = (1 - probability_abort) * usable_pulses_lp / self.number_of_pulses

        self.last_positive = key_rate
        self.last_x = x
        return key_rate

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        """Extract the parameters and assigns them correspondingly."""
        return {
            "mu": np.array(x[0 : self.number_of_decoy + 1]),
            "probability_basis_X": np.array(
                x[self.number_of_decoy + 1 : 2 * self.number_of_decoy + 2]
            ),
            "probability_basis_Z": np.array(x[2 * self.number_of_decoy + 2 :]),
        }

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        The laser intensities should be ordered and the probabilities should
        sum to one. Probabilities for both X and Z-basis are considered
        simultaneously. We consider the Z-basis for error estimation and the
        X-basis for key-rate estimation, so no sifting rate is considered.

        For certain input parameters it can happen that the resulting lp problem is
        unfeasible. In that case the attenuation is slightly modified (``+1e-8``) in an
        attempt to obtain a feasible lp problem that can be solved.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value
            bounds: Bounds on search range
            args: Other variables to be optimized, for instance the attenuation

        Returns:
            Optimized x=[intensity, probability_basis_X, probability_basis_Z]
            and found optimal key-rate

        Raises:
            ValueError: When x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
            OptimizationError: When lp solver is unsuccessful due to infeasible problem.
                Multiple attempts are made with slightly modified  attenuation before
                error is raised.
        """
        less_one = 1 - 1e-16

        if bounds is None:
            lower_bound = np.hstack(
                (
                    1e-16 * np.ones((self.number_of_decoy + 1)),
                    np.zeros(2 * self.number_of_decoy + 2),
                )
            )
            upper_bound = np.hstack(
                (
                    np.ones((self.number_of_decoy + 1)),
                    np.ones(2 * self.number_of_decoy + 2),
                )
            )
        else:
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
            if len(bounds) != 3 * (self.number_of_decoy + 1):
                error_msg = (
                    "Invalid dimensions input bounds. Expected "
                    f"{3 * (self.number_of_decoy + 1)} upper- and lower bounds."
                    f"Received {len(bounds)} bounds."
                )
                raise ValueError(error_msg)
        optimize_bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        if x0 is None:
            x0 = (lower_bound + upper_bound) / 2
        x0 = np.asarray(x0)
        if len(x0) != 3 * (self.number_of_decoy + 1):
            error_msg = (
                "Invalid number of inputs. Expected "
                f"{3 * (self.number_of_decoy + 1)} inputs. Received "
                f"{len(x0)} inputs."
            )
            raise ValueError(error_msg)

        # The probabilities are normalized to 1
        x0[self.number_of_decoy + 1 :] /= x0[self.number_of_decoy + 1 :].sum()

        self.last_positive = -1
        self.last_x = x0

        # The constraints enforce the intensities to be decreasing
        # Furthermore, it enforces that the probabilities are between zero and
        # one and that they together sum to one
        b = np.hstack(
            (
                np.repeat(-np.inf, self.number_of_decoy),
                np.zeros(self.number_of_decoy + 1),
                1,
            )
        )
        A = np.vstack(
            (
                np.flip(
                    np.hstack(
                        (
                            (
                                np.triu(
                                    np.ones(
                                        (
                                            self.number_of_decoy,
                                            self.number_of_decoy + 1,
                                        ),
                                        dtype=int,
                                    ),
                                    1,
                                )
                                - np.hstack(
                                    (
                                        np.diag(
                                            less_one
                                            * np.ones(self.number_of_decoy, dtype=int)
                                        ),
                                        np.zeros((self.number_of_decoy, 1), dtype=int),
                                    )
                                )
                            ),
                            np.zeros(
                                (self.number_of_decoy, 2 * self.number_of_decoy + 2),
                                dtype=int,
                            ),
                        )
                    ),
                    0,
                ),
                np.eye(self.number_of_decoy + 1, 3 * self.number_of_decoy + 3),
                np.hstack(
                    (
                        np.zeros(self.number_of_decoy + 1, dtype=int),
                        np.ones(2 * self.number_of_decoy + 2, dtype=int),
                    )
                ),
            )
        )
        B = np.hstack(
            (
                np.zeros(self.number_of_decoy, dtype=int),
                np.ones(self.number_of_decoy + 2),
            )
        )

        constraint = scipy.optimize.LinearConstraint(A, b, B)
        # In some cases the gradient appears to be zero, indicating a linear function.
        # It is not a linear function, hence we suppress the error
        args = {"attenuation": attenuation}
        num_attempts = 3
        for _ in range(num_attempts):  # Maximum 3 retries
            try:
                res = scipy.optimize.minimize(
                    self._f,
                    x0,
                    args=args,
                    constraints=(constraint),
                    bounds=optimize_bounds,
                    **NLP_CONFIG,
                )
            except OptimizationError:  # noqa: PERF203 Redo calculation
                args["attenuation"] += 1e-8
            else:
                rate = np.atleast_1d(-res.fun)[0]
                if rate < 0:  # Retry when negative key-rate is found.
                    args["attenuation"] += 1e-8
                    continue
                return self._extract_parameters(res.x), float(rate)
        error_msg = "Unable to find solution for optimal key rate."
        raise OptimizationError(error_msg)



"""Base class for error corrector objects."""

from __future__ import annotations

import abc
import hashlib
import hmac
from dataclasses import dataclass, fields
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.classical import (
        Message,
        ReceiverBase,
        SenderBase,
    )


@dataclass
class CorrectorOutputBase:
    """Base class corrector summary object.

    Args:
        input_alice: Input message Alice
        output_alice: Corrected message Alice
        input_bob: Input message Bob
        output_bob: Corrected message Bob
        input_error: Input error rate
        output_error: Output error rate
        output_length: Output message length
        number_of_exposed_bits: Number of bits exposed in protocol
        key_reconciliation_rate: Key reconciliation efficiency
        number_of_communication_rounds: Number of communication rounds
    """

    input_alice: Message
    output_alice: Message
    input_bob: Message
    output_bob: Message
    input_error: float
    output_error: float
    output_length: int
    number_of_exposed_bits: int
    key_reconciliation_rate: float
    number_of_communication_rounds: int

    def __str__(self) -> str:
        """String representation of CorrectorOutputBase."""
        res = "Corrector summary:"
        for field in fields(self):
            res += f"\n{field.name} ({field.type}):\t {getattr(self, field.name)}"
        return res


class Corrector(metaclass=abc.ABCMeta):
    """Error corrector base class."""

    def __init__(
        self,
        alice: SenderBase,
        bob: ReceiverBase,
    ) -> None:
        """Base class for error correcting.

        Args:
            alice: The sending party
            bob: The receiving party
        """
        self.alice = alice
        self.bob = bob

    def correct_errors(self, detail_transcript: bool = False) -> CorrectorOutputBase:  # noqa: FBT001, FBT002
        """Receiver Bob corrects the errors based on Alice her message.

        Args:
            detail_transcript: Whether to print a detailed transcript
        """
        self.bob.correct_errors(self.alice)

        if detail_transcript:
            print(self.bob.transcript)  # noqa: T201

        return self.summary()

    @abc.abstractmethod
    def summary(self) -> CorrectorOutputBase:
        """Calculate a summary object for the error correction containing.

        Summary contains:

            - original message
            - corrected message
            - error rate (before and after correction)
            - number_of_exposed_bits
            - key_reconciliation_rate
            - protocol specific parameters
        """

    @staticmethod
    def calculate_number_of_errors(message1: Message, message2: Message) -> int:
        """Calculate the error rate between two messages.

        If messages differ in length, the number of errors is calculated
        using the number of bits of the shortest message.

        Args:
            message1: First message
            message2: Second message

        Returns:
            number_of_errors: Number of errors.

        Raises:
            ValueError: If either message has length zero.

        """
        if message1.length == 0 or message2.length == 0:
            error_msg = "One of the messages has length zero."
            raise ValueError(error_msg)
        return sum(x != y for (x, y) in zip(message1.message, message2.message))

    @staticmethod
    def calculate_error_rate(message1: Message, message2: Message) -> float:
        """Calculate the error rate between two messages.

        If messages differ in length, the number of errors is calculated
        using the number of bits of the shortest message.

        Args:
            message1: First message
            message2: Second message

        Returns:
            error_rate: Ratio of errors over the message length.
        """
        return Corrector.calculate_number_of_errors(message1, message2) / min(
            message1.length, message2.length
        )

    def calculate_key_reconciliation_rate(self, *, exposed_bits: bool = False) -> float:
        """Calculate the key reconciliation rate.

        Args:
            exposed_bits: If true, uses the number of exposed bits to compute the
            key-reconciliation rate. Otherwise, uses the ratio between the in- and
            output message length.

        Returns:
            key_rate: The reconciliation rate
        """
        if exposed_bits:
            key_rate = (
                self.alice.message.length - self.alice.net_exposed_bits
            ) / self.alice.original_message.length
        else:
            key_rate = self.alice.message.length / self.alice.original_message.length
        return key_rate

    @staticmethod
    def create_message_tag_pair(
        message: Message, shared_key: str
    ) -> tuple[bytes, bytes]:
        """Prepares a message-tag hashed pair.

        The message can be communicated publicly. The tag is the hash of the message,
        given a key.

        Args:
            message: To be communicated message
            shared_key: Shared secret key

        Returns:
            message: To be communicated message
            tag: Hash of the message, given the key, with length of the key
        """
        message_str = "".join(str(x) for x in message.message)
        shared_key_bytes = bytes(shared_key.encode("utf-8"))
        message_bytes = bytes(message_str.encode("utf-8"))

        tag = hmac.new(
            key=shared_key_bytes, msg=message_bytes, digestmod=hashlib.sha384
        )
        return message_bytes, tag.digest()



"""Base config settings."""

from scipy import optimize

# Options to be used for solving the linear programs
LP_CONFIG = {
    "method": "highs",
    "options": {"maxiter": 1000, "disp": False, "presolve": True},
}

NLP_CONFIG = {
    "method": "trust-constr",
    "jac": "2-point",
    "hess": optimize.BFGS(),
    "options": {"maxiter": 1000, "disp": False, "finite_diff_rel_step": 1e-6},
}


class OptimizationError(ValueError):
    """Raised when optimization is unsuccessful.

    This error typically thrown when the lp problem is infeasible.
    """



"""Classes to perform a Winnow error correction protocol.

The Winnow error correction protocol is based on Hamming codes. An advantage of the protocol
is that it requires less communication than other error correction protocols. The protocol
however might introduce errors in specific cases. With every communication, the Winnow protocol
leaks information to potential eavesdroppers. This can be overcome by discarding message bits
equal to the amount of information leaks, thereby achieving privacy maintenance.

Typical usage example:

    >>> from tno.quantum.communication.qkd_key_rate.classical import (
    ...     Message,
    ...     Schedule,
    ...     Permutations,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.classical.winnow import (
    ...     WinnowCorrector,
    ...     WinnowReceiver,
    ...     WinnowSender,
    ... )
    >>>
    >>> message_length = 100_000
    >>> error_rate = 0.05
    >>> input_message = Message.random_message(message_length=message_length, random_state=43)
    >>> error_message = input_message.apply_errors(error_rate=0.05, random_state=42)
    >>>
    >>> schedule = Schedule.schedule_from_error_rate(error_rate=error_rate)
    >>> number_of_passes = sum(schedule.schedule)
    >>> permutations = Permutations.random_permutation(
    ...     number_of_passes=number_of_passes, message_size=message_length, random_state=42
    ... )
    >>>
    >>> alice = WinnowSender(
    ...     message=input_message, permutations=permutations, schedule=schedule
    ... )
    >>> bob = WinnowReceiver(
    ...     message=error_message, permutations=permutations, schedule=schedule
    ... )
    >>> corrector = WinnowCorrector(alice=alice, bob=bob)
    >>>
    >>> print(corrector.correct_errors())  # doctest: +SKIP
    Corrector summary:
    input_alice (Message):   00111001111000101101111101011000100010000100000011...
    output_alice (Message):  01111111100100100011010100110110101100010110101111...
    input_bob (Message):     00111001110000101101111101011100100010000110000011...
    output_bob (Message):    01111111100100100011010100110110101100010110101111...
    input_error (float):     0.04959
    output_error (float):    0.0
    output_length (int):     52063
    number_of_exposed_bits (int):    3647
    key_reconciliation_rate (float):         0.48416
    number_of_communication_rounds (int):    15
    schedule (list[int]):    [2, 1, 1, 0, 0, 1, 1, 1]
"""  # noqa: E501

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical import (
    Corrector,
    CorrectorOutputBase,
    Permutations,
    ReceiverBase,
    Schedule,
    SenderBase,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.classical import Message


class WinnowSender(SenderBase):
    """This class encodes all functions available to both sender and receiver.

    It keeps track of the number of exposed bits and can compute syndromes and
    parities. Furthermore, it keeps track of the blocks with errors
    """

    def __init__(
        self,
        message: Message,
        permutations: Permutations,
        schedule: Schedule,
        name: str | None = None,
    ) -> None:
        """Init of a WinnowSender.

        Args:
            message: Input message of the sender party
            permutations: List containing permutations for each pass
            schedule: The winnow schedule
            name: Name of the sender party
        """
        super().__init__(message=message, permutations=permutations, name=name)
        self.schedule = schedule

        self.number_of_bad_blocks = 0
        self.block_size = 8
        self.number_of_blocks = 0
        self.maximum_number_of_communication_rounds = 0
        self._number_of_exposed_bits = 0
        self.number_of_passes = permutations.number_of_passes
        self._net_exposed_bits = 0
        self.syndrome_length = 3
        self.removed_bits: list[int] = []

        self.syndrome_array: list[int] = [0 for _ in range(self.message.length // 8)]

        # Stores the indices of blocks which contain errors
        self.bad_blocks_array: list[int] = [0 for _ in range(self.message.length // 8)]
        self.parity_string: list[int] = []

        self.parity_check_matrix = np.zeros([10, 1023], dtype=int)

        self.transcript = ""

        self.first_pass()

    def create_parity_check_matrix(self) -> None:
        """Creates a parity check matrix.

        This matrix is used to encode the bit strings.
        """
        size = 1 << self.syndrome_length
        for i in range(self.syndrome_length):
            for j in range(1, size):
                self.parity_check_matrix[i, j - 1] = int(j / (1 << i)) & 0x1

    def get_parity(self, index_start: int, index_end: int) -> int:
        """Get the parity of a specific message part between two indices.

        Args:
            index_start: Start index of the message
            index_end: End index of the message

        Returns:
            Parity of substring
        """
        number_of_ones = 0
        for i in range(index_start, index_end):
            if self.message[i] == 1:
                number_of_ones += 1
        return number_of_ones % 2

    def build_parity_string(self) -> None:
        """Builds a parity string for all blocks."""
        for i in range(self.number_of_blocks):
            index_start = i * self.block_size
            index_end = index_start + self.block_size

            parity = self.get_parity(index_start, index_end)
            self.parity_string[i] = parity

            self._number_of_exposed_bits += 1
            self._net_exposed_bits += 1

    def discard_parity_bits(self) -> None:
        """The first bit of every parity block is discarded."""
        old_index = 0
        counter = -1
        while old_index < self.message.length:
            if ((old_index % self.block_size) == 0) and (
                old_index != (self.number_of_blocks * self.block_size)
            ):
                # Is it the first bit of a block and is it not the last block
                counter += 1
                self._net_exposed_bits -= 1
                self.message.pop(old_index - counter)
            old_index += 1

        self.block_size -= 1
        self.transcript += (
            f"\tBoth discard the parity bits for pass {self.schedule.pass_number}.\n"
        )

    def get_syndrome(self, index_block: int) -> int:
        """Computes the syndrome of a block.

        Both parties compute their syndrome individually, hence, no
        communication is needed here.
        """
        if index_block > self.number_of_blocks:
            self.transcript += (
                "Illegal block number. Returning block_size + 1 for new syndrome.\n"
            )
            return self.block_size + 1

        placeholder = 0
        new_syndrome = 0
        # Computer he highest order bit of the syndrome first and then work down
        for i in range(self.syndrome_length - 1, -1, -1):
            new_syndrome <<= 1
            # Multiply the block by the (i-1)-th row of the parity check matrix
            # and add it to the syndrome
            for j in range(self.block_size):
                placeholder += (
                    self.parity_check_matrix[i, j]
                    * self.message[index_block * self.block_size + j]
                )
                placeholder &= 0x1
            new_syndrome += placeholder
            placeholder = 0

        self._number_of_exposed_bits += self.syndrome_length
        self._net_exposed_bits += self.syndrome_length
        return new_syndrome

    def disagreeing_block_parities(self, alice: WinnowSender) -> None:
        """Finds the disagreeing block parities.

        The found parities of both parties are compared. This can be done with
        two communication rounds (one both ways).
        Afterwards, both separately process the results.

        Args:
            alice: The sending party
        """
        self.maximum_number_of_communication_rounds += 1
        self.transcript += (
            f"Bob sends his parity string to alice for pass {self.schedule.pass_number}"
            ". Alice compares the string with hers and keeps track of the disagreeing"
            " blocks.\n"
        )

        counter_for_bad_blocks = 0
        for i in range(self.number_of_blocks):
            if alice.parity_string[i] != self.parity_string[i]:
                # The parities disagree, save the block-index as bad block.
                self.bad_blocks_array[counter_for_bad_blocks] = i
                alice.bad_blocks_array[counter_for_bad_blocks] = i
                counter_for_bad_blocks += 1
        self.number_of_bad_blocks = counter_for_bad_blocks
        alice.number_of_bad_blocks = counter_for_bad_blocks

    def discard_syndrome_bits(self) -> None:
        """Discards syndrome bits.

        Bits at indices $2^j-1$ are removed. These correspond to the linearly
        independent columns of the parity check matrix.

        In this function the number of bad blocks is known.

        No communication is needed to discard syndrome bits.
        """
        counter_for_error_blocks = 0
        removed_bits = []
        for index_block in range(self.number_of_blocks):
            # If we have not hit all bad blocks, and the counter is at a bad block
            if (
                counter_for_error_blocks < self.number_of_bad_blocks
                and self.bad_blocks_array[counter_for_error_blocks] == index_block
            ):
                power = 0
                counter_for_error_blocks += 1
                offset_counter = -1
                for index_bit in range(self.block_size):
                    if (index_bit + 1) == (1 << power):
                        # Discard bits if they are at a location with index a
                        # power of 2 - 1
                        power += 1
                        offset_counter += 1
                        self._net_exposed_bits -= 1
                        removed_bits.append(index_bit - offset_counter)
                        self.message.pop(index_bit - offset_counter)

        self.transcript += (
            f"\tBoth discard the syndrome bits for pass {self.schedule.pass_number}.\n"
        )

    def build_syndrome_string(self, alice: WinnowSender) -> None:
        """Create a syndrome string for all disagreeing blocks.

        Computes the syndrome for blocks with disagreeing parity.

        Args:
            alice: The sending party
        """
        self.disagreeing_block_parities(alice)

        for i in range(self.number_of_bad_blocks):
            alice.syndrome_array[i] = self.get_syndrome(alice.bad_blocks_array[i])

    def first_pass(self) -> None:
        """First pass with initializations and parity determination."""
        # Note that the incomplete last block is not included in each pass
        self.number_of_blocks = self.message.length // self.block_size
        self.parity_string = [0] * self.number_of_blocks

        self.create_parity_check_matrix()
        self.permute_buffer()
        self.build_parity_string()
        self.discard_parity_bits()

    def next_pass(self) -> None:
        """Performs computations to prepare for the next pass.

        Computations include permuting the message and creating a new parity string.
        """
        i = self.schedule.next_pass()
        self.syndrome_length = i + 3
        self.block_size = 1 << self.syndrome_length

        # Note that the incomplete last block is not included in each pass
        self.number_of_blocks = self.message.length // self.block_size
        self.create_parity_check_matrix()
        self.permute_buffer()
        self.build_parity_string()
        self.discard_parity_bits()

    def permute_buffer(self) -> None:
        """Permutes the message string."""
        self.permutations.shorten_pass(self.schedule.pass_number, self.message.length)
        self.message.apply_permutation(self.permutations[self.schedule.pass_number])


class WinnowReceiver(WinnowSender, ReceiverBase):
    """This class encodes all functions only available to the receiver.

    The receiver is assumed to have a string with errors and is thus assumed to
    correct the errors.
    """

    def __init__(
        self,
        message: Message,
        permutations: Permutations,
        schedule: Schedule,
        name: str | None = None,
    ) -> None:
        """Init of WinnowReceiver.

        Args:
            message: Input message of the sender party
            permutations: List containing permutations for each pass
            schedule: The winnow schedule.
            name: Name of the receiver party
        """
        super().__init__(
            message, name=name, permutations=permutations, schedule=schedule
        )

    def fix_errors_with_syndrome(self, alice: WinnowSender) -> None:
        """Corrects errors using the syndrome strings of alice and bob.

        Args:
            alice: The sending party
        """
        syndrome_alice = alice.syndrome_array

        for i in range(alice.number_of_bad_blocks):
            syndrome_bob = alice.get_syndrome(alice.bad_blocks_array[i])
            disagreeing_syndrome_bit = (
                syndrome_alice[i] ^ syndrome_bob
            )  # XOR the two syndromes

            if disagreeing_syndrome_bit == 0:
                # Erroneous bit was already discarded
                continue
            self.correct_individual_error(
                alice.bad_blocks_array[i] * self.block_size
                + disagreeing_syndrome_bit
                - 1
            )
        self.transcript += "bob computes the syndromes for his bit string and \
accordingly corrects bits in his own bit string based on the difference with \
the syndromes of alice.\n"

        self.discard_syndrome_bits()
        alice.discard_syndrome_bits()

    def correct_errors(self, alice: SenderBase) -> None:
        """The main routine, finds all errors and corrects them.

        It is assumed that Alice and Bob use one communication round to agree on
        the used permutations. Afterwards, they use two communication rounds per
        iteration to communicate the syndromes.

        Args:
            alice: The sending party

        Raises:
            TypeError: If alice is not a WinnowSender.
        """
        if not isinstance(alice, WinnowSender):
            error_msg = "alice is not of type `WinnowSender`."
            raise TypeError(error_msg)
        alice.schedule = deepcopy(self.schedule)

        self.maximum_number_of_communication_rounds += 1
        if (len(self.schedule) - len(self.permutations)) >= 0:
            # Add permutations, if there are not enough for the whole schedule
            self.permutations += Permutations.random_permutation(
                number_of_passes=len(self.schedule) - len(self.permutations) + 1,
                message_size=self.message.length,
            )
            self.transcript += (
                "Permutations for bit string shared between alice and bob.\n"
            )
            alice.permutations = deepcopy(self.permutations)

        number_of_remaining_passes = 1

        while number_of_remaining_passes > 0:
            self.maximum_number_of_communication_rounds += 1
            self.transcript += (
                "alice sends the syndromes for the disagreeing blocks of pass"
                f"{self.schedule.pass_number} to bob, as well as the block indices."
                "If there are no disagreeing blocks, both enter the next pass.\n"
            )
            self.build_syndrome_string(alice)

            # If we have blocks with disagreeing syndromes, correct bits accordingly
            if self.number_of_bad_blocks != 0:
                self.fix_errors_with_syndrome(alice)
            else:
                self.transcript += (
                    f"\tNo disagreeing parities in pass {self.schedule.pass_number}"
                    ", continue.\n"
                )

            self.next_pass()
            alice.next_pass()

            number_of_remaining_passes = self.schedule.remaining_passes


@dataclass
class WinnowCorrectorOutput(CorrectorOutputBase):
    """Data class for Winnow Corrector output."""

    schedule: Schedule


class WinnowCorrector(Corrector):
    """Winnow corrector."""

    def __init__(self, alice: WinnowSender, bob: WinnowReceiver) -> None:
        """Init of WinnowCorrector.

        Args:
            alice: Sender party.
            bob: Receiver party.

        Raises:
            ValueError: If permutations alice are not same as permutations bob.
        """
        self.alice: WinnowSender = alice
        self.bob: WinnowReceiver = bob

        if self.alice.permutations != self.bob.permutations:
            error_msg = "Permutations Alice not same as permutations Bob."
            raise ValueError(error_msg)

    def summary(self) -> WinnowCorrectorOutput:
        """Calculate a summary object for the error correction.

        Summary contains
            - original message
            - corrected message
            - error rate (before and after correction)
            - number_of_exposed_bits
            - key_reconciliation_rate
            - protocol specific parameters
        """
        return WinnowCorrectorOutput(
            input_alice=self.alice.original_message,
            output_alice=self.alice.message,
            input_bob=self.bob.original_message,
            output_bob=self.bob.message,
            input_error=self.calculate_error_rate(
                self.alice.original_message, self.bob.original_message
            ),
            output_error=self.calculate_error_rate(
                self.alice.message, self.bob.message
            ),
            output_length=self.alice.message.length,
            number_of_exposed_bits=self.bob.net_exposed_bits,
            key_reconciliation_rate=self.calculate_key_reconciliation_rate(
                exposed_bits=True
            ),
            number_of_communication_rounds=self.bob.maximum_number_of_communication_rounds,
            schedule=self.bob.schedule,
        )



"""Base class for detector objects."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from tno.quantum.utils.validation import check_path

if TYPE_CHECKING:
    from pathlib import Path

DEFAULT_POLARIZATION_DRIFT = np.arcsin(np.sqrt(0.015))
DEFAULT_ERROR_DETECTOR = 0


class Detector:
    """Class for the detectors to be used in the key-rate estimates."""

    required_fields = (
        "name",
        "jitter_source",
        "jitter_detector",
        "dead_time",
        "detection_window",
        "efficiency_system",
    )

    def __init__(  # noqa: PLR0912, PLR0913, C901, PLR0915
        self,
        name: str,
        jitter_source: float,
        jitter_detector: float,
        dead_time: float,
        detection_window: float,
        efficiency_system: float,
        polarization_drift: float = DEFAULT_POLARIZATION_DRIFT,
        error_detector: float = DEFAULT_ERROR_DETECTOR,
        dark_count_frequency: float | None = None,
        dark_count_rate: float | None = None,
        detection_frequency: float | None = None,
        interval: float | None = None,
        efficiency_detector: float | None = None,
        efficiency_party: float | None = None,
    ) -> None:
        r"""Initialise a Detector instance.

        Args:
            name: Label of the detector party
            jitter_source: Time-delay introduced at the source
            jitter_detector: Time-delay introduced at the detector
            dead_time: The recovery time before a new event can be recorded
            detection_window: Time window used for event detection
            efficiency_system: Efficiency of the detecting side without the detector
            polarization_drift: Shift/drift in the encoding of the photons
            error_detector: Error-rate of detector
            dark_count_frequency: Number of dark-counts per second
            dark_count_rate: Number of dark-counts per detection window
            detection_frequency: Number of detection windows per second
            interval: Length of a single detection window in seconds
            efficiency_detector: Efficiency of the detector on the detecting side
            efficiency_party: Total efficiency of the detecting side

        If related optional arguments are both given they must satisfy

            - $\text{interval} = \frac{1}{\text{detection_frequency}}$
            - $\text{dark_count_frequency} = \text{dark_count_rate} \cdot \text{detection_frequency}$
            - $\text{efficiency_party} = \text{efficiency_system} \cdot \text{efficiency_detector}$

        If they are not given, they are calculated using the same formulas.

        Returns:
            Detector instance with specified parameters.

        Raises:
            ValueError: If a required field is missing.
            ValueError: If a needed optional field is missing.
            ValueError: If inconsistent related optional fields are provided.
        """  # noqa: E501
        self.name = name
        self.jitter_source = jitter_source
        self.jitter_detector = jitter_detector
        self.dead_time = dead_time
        self.detection_window = detection_window
        self.efficiency_system = efficiency_system
        self.polarization_drift = polarization_drift
        self.error_detector = error_detector

        for k in self.required_fields:
            if self.__dict__[k] is None:
                error_msg = f"Field '{k}' is required field."
                raise ValueError(error_msg)

        # Handle optional fields
        tol = 10e-6

        # section: interval and detection frequency
        if interval is None and detection_frequency is None:
            error_msg = (
                "Either the field 'interval' or 'detection_frequency' is required."
            )
            raise ValueError(error_msg)
        if interval is not None and detection_frequency is not None:
            self.interval = interval
            self.detection_frequency = detection_frequency
            if not abs(interval - 1 / detection_frequency) < tol:
                error_msg = "Incompatible interval and detection frequency."
                raise ValueError(error_msg)
        if interval is None and detection_frequency is not None:
            self.detection_frequency = detection_frequency
            self.interval = 1 / self.detection_frequency
        if detection_frequency is None and interval is not None:
            self.interval = interval
            self.detection_frequency = 1 / self.interval

        # section: dark count rate and dark count frequency
        if dark_count_rate is None and dark_count_frequency is None:
            error_msg = (
                "Either the field 'dark_count_rate' or "
                "'dark_count_frequency' is required."
            )
            raise ValueError(error_msg)
        if dark_count_rate is not None and dark_count_frequency is not None:
            self.dark_count_rate = dark_count_rate
            self.dark_count_frequency = dark_count_frequency
            if not (dark_count_rate - dark_count_frequency * self.interval) < tol:
                error_msg = "Incompatible dark count rate and dark count frequency."
                raise ValueError(error_msg)

        if dark_count_rate is None and dark_count_frequency is not None:
            self.dark_count_frequency = dark_count_frequency
            self.dark_count_rate = self.dark_count_frequency * self.interval
        if dark_count_frequency is None and dark_count_rate is not None:
            self.dark_count_rate = dark_count_rate
            self.dark_count_frequency = dark_count_rate * self.detection_frequency

        # section: efficiency party and efficiency detector
        if efficiency_party is None and efficiency_detector is None:
            error_msg = (
                "Either the field 'efficiency_party' or "
                "'efficiency_detector' is required."
            )
            raise ValueError(error_msg)
        if efficiency_party is not None and efficiency_detector is not None:
            self.efficiency_party = efficiency_party
            self.efficiency_detector = efficiency_detector
            if not (efficiency_party - efficiency_system * efficiency_detector) < tol:
                error_msg = "Incompatible efficiency and efficiency detector."
                raise ValueError(error_msg)
        if efficiency_party is None and efficiency_detector is not None:
            self.efficiency_detector = efficiency_detector
            self.efficiency_party = efficiency_system * self.efficiency_detector
        if efficiency_detector is None and efficiency_party is not None:
            self.efficiency_party = efficiency_party
            self.efficiency_detector = efficiency_party / efficiency_system

    @classmethod
    def from_file(cls, path: str | Path) -> list[Detector]:
        """Construct Detectors from csv file.

        Args:
            path: Path to csv file.
        """
        path = check_path(
            path, "path", must_exist=True, must_be_file=True, required_suffix=".csv"
        )
        detector = pd.read_csv(path, delimiter=";")
        detector = detector.replace({np.nan: None})
        return [
            cls(**detector_kwargs) for detector_kwargs in detector.to_dict("records")
        ]

    def customise(  # noqa: PLR0913, C901
        self,
        name: str | None = None,
        jitter_source: float | None = None,
        jitter_detector: float | None = None,
        dead_time: float | None = None,
        detection_window: float | None = None,
        efficiency_system: float | None = None,
        polarization_drift: float | None = None,
        error_detector: float | None = None,
        dark_count_frequency: float | None = None,
        dark_count_rate: float | None = None,
        detection_frequency: float | None = None,
        interval: float | None = None,
        efficiency_detector: float | None = None,
        efficiency_party: float | None = None,
    ) -> Detector:
        """Create a detector with customised parameter from current detector.

        See :py:meth:`~__init__` for parameter description.
        """
        new_parameters = dict(**self.__dict__)
        if jitter_source is not None:
            new_parameters.update(jitter_source=jitter_source)
        if jitter_detector is not None:
            new_parameters.update(jitter_detector=jitter_detector)
        if dead_time is not None:
            new_parameters.update(dead_time=dead_time)
        if detection_window is not None:
            new_parameters.update(detection_window=detection_window)
        if efficiency_system is not None:
            new_parameters.update(efficiency_system=efficiency_system)
        if polarization_drift is not None:
            new_parameters.update(polarization_drift=polarization_drift)
        if error_detector is not None:
            new_parameters.update(error_detector=error_detector)

        if dark_count_frequency is not None or dark_count_rate is not None:
            new_parameters.update(dark_count_frequency=dark_count_frequency)
            new_parameters.update(dark_count_rate=dark_count_rate)

        if detection_frequency is not None or interval is not None:
            new_parameters.update(detection_frequency=detection_frequency)
            new_parameters.update(interval=interval)

        if efficiency_detector is not None or efficiency_party is not None:
            new_parameters.update(efficiency_detector=efficiency_detector)
            new_parameters.update(efficiency_party=efficiency_party)

        new_name = name if name is not None else self.name + "(adjusted)"
        new_parameters.update(name=new_name)

        return Detector(**new_parameters)

    def get_parameters(self) -> dict[str, Any]:
        """Get all parameters of the Detector."""
        return dict(**self.__dict__)

    def __repr__(self) -> str:
        """Repr of a detector."""
        return f"Detector:{self.name}"


standard_detector = Detector(
    name="standard",
    efficiency_detector=0.2,
    jitter_source=0,
    jitter_detector=5.00e-11,
    dead_time=4.50e-08,
    dark_count_frequency=100,
    detection_frequency=1.00e07,
    detection_window=5,
    efficiency_system=1,
)



"""Tests for BB84 single photon."""

import numpy as np

from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
from tno.quantum.communication.qkd_key_rate.quantum.bb84_single_photon import (
    BB84SingleAsymptoticKeyRateEstimate,
)


def test_zero_dark_count() -> None:
    detector = standard_detector.customise(dark_count_rate=0)

    distance = np.arange(1, 300)  # Distance between ALice and Bob in km
    attenuation_factor = 0.2  # Channel loss in dB per km
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    x0 = None

    asymptotic_key_rate = BB84SingleAsymptoticKeyRateEstimate(detector=detector)

    for i, att in enumerate(attenuation):
        x, rate = asymptotic_key_rate.optimize_rate(
            x0=x0, bounds=None, attenuation=float(att)
        )
        key_rate[i] = rate
        x0 = x["mu"]

    assert all(rate > 0 for rate in key_rate)


def test_single_photon() -> None:
    distance = np.arange(1, 50)  # Distance between ALice and Bob in km
    attenuation_factor = 0.2  # Channel loss in dB per km
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    x0 = None

    asymptotic_key_rate = BB84SingleAsymptoticKeyRateEstimate(
        detector=standard_detector
    )

    for i, att in enumerate(attenuation):
        x, rate = asymptotic_key_rate.optimize_rate(
            x0=x0, bounds=None, attenuation=float(att)
        )
        key_rate[i] = rate
        x0 = x["mu"]

    assert all(rate > 0 for rate in key_rate)



"""Test the ParityStrategy base object."""

import numpy as np
import pytest

from tno.quantum.communication.qkd_key_rate.classical import ParityStrategy


@pytest.mark.parametrize(
    ("error_rate", "expected_block_size"),
    [
        (0.009, 24),
        (0.019, 18),
        (0.029, 12),
        (0.039, 11),
        (0.049, 10),
        (0.059, 9),
        (0.069, 7),
        (0.079, 7),
        (0.089, 7),
        (0.099, 5),
        (0.109, 5),
        (0.119, 5),
        (0.129, 4),
        (0.139, 4),
        (0.149, 4),
    ],
)
def test_get_start_block_size(error_rate: float, expected_block_size: int) -> None:
    """
    Test get_start_block_size function
    """
    parity_strategy = ParityStrategy(error_rate=error_rate)
    start_block_size = parity_strategy.get_start_block_size()
    assert start_block_size == expected_block_size


def test_stop_high_error() -> None:
    """
    Test to high error rate
    """
    error_rate = 0.16
    parity_strategy = ParityStrategy(error_rate=error_rate)

    expected_msg = "Error rate too high for secure protocol"
    with pytest.raises(ValueError, match=expected_msg):
        parity_strategy.get_start_block_size()


def test_calculate_message_parity_strategy() -> None:
    """Test calculate message parity strategy"""
    error_rate = 0.149
    number_of_passes = 10
    switch_after_pass = 5
    sampling_fraction = 0.34

    parity_strategy = ParityStrategy(
        error_rate=error_rate,
        number_of_passes=number_of_passes,
        sampling_fraction=sampling_fraction,
        switch_after_pass=switch_after_pass,
    )

    message_size = 1024
    size_blocks_parities = parity_strategy.calculate_message_parity_strategy(
        message_size
    )

    # Normal passes
    expected_block_size, expected_number_of_blocks = 4, message_size / 4
    for block_size, number_of_blocks in size_blocks_parities[0:switch_after_pass]:
        assert block_size == expected_block_size
        assert number_of_blocks == expected_number_of_blocks

        expected_block_size *= 2
        expected_number_of_blocks /= 2

    # Random permutations
    expected_block_size = np.ceil(message_size * sampling_fraction)
    expected_number_of_blocks = np.ceil(message_size / expected_block_size)
    for block_size, number_of_blocks in size_blocks_parities[switch_after_pass:]:
        assert block_size == expected_block_size
        assert number_of_blocks == expected_number_of_blocks



"""Base class for ParityStrategy object."""

from __future__ import annotations

import numpy as np

ERROR_RATES_TO_BLOCK_SIZES = {
    0.010: 24,
    0.020: 18,
    0.030: 12,
    0.040: 11,
    0.050: 10,
    0.060: 9,
    0.070: 7,
    0.080: 7,
    0.090: 7,
    0.100: 5,
    0.110: 5,
    0.120: 5,
    0.130: 4,
    0.140: 4,
    0.150: 4,
}


class ParityStrategy:
    """Parity strategy to be used in the Cascade protocol.

    In the Cascade protocol, for efficiency reasons, we may change the strategy
    of our blocks after some passes. This class deals with the parity strategy
    """

    def __init__(
        self,
        error_rate: float,
        sampling_fraction: float = 0.5,
        number_of_passes: int = 10,
        switch_after_pass: int | None = None,
    ) -> None:
        """Init of ParityStrategy.

        Args:
            error_rate: An estimate of the error rate in the message
            sampling_fraction: Fraction of the string that is sampled after
            switching strategies
            number_of_passes: Total number of Cascade iterations
            switch_after_pass: After which pass we switch the strategy
        """
        self.error_rate = error_rate
        self.number_of_passes = number_of_passes
        self.switch_after_pass = switch_after_pass
        self.sampling_fraction = sampling_fraction

    def get_start_block_size(self) -> int:
        """Determine starting block size.

        Returns:
            The largest block size that is compatible with given error-rate

        Raises:
            ValueError: If error rate is to high for a secure protocol.
        """
        possible_block_sizes = [
            block_size
            for (error_rate, block_size) in ERROR_RATES_TO_BLOCK_SIZES.items()
            if error_rate > self.error_rate
        ]

        if possible_block_sizes == []:
            error_msg = "Error rate too high for secure protocol"
            raise ValueError(error_msg)
        return max(possible_block_sizes)

    def calculate_message_parity_strategy(
        self, message_size: int
    ) -> tuple[tuple[int, int], ...]:
        """Sets the parity strategy.

        Initial strategy is to double the block size in each subsequent pass.
        After some number of passes, we may change the strategy by randomly
        sampling bits from the message in new blocks.

        Args:
            message_size: Size of the message for which to calculate the parity strategy

        Returns:
            size_blocks_parities: The block size and number of blocks for each pass
        """
        if self.switch_after_pass is None:
            self.switch_after_pass = self.number_of_passes
        start_block_size = self.get_start_block_size()

        # First passes using classical cascade
        block_sizes = (
            start_block_size << index_pass
            for index_pass in range(self.switch_after_pass)
        )
        size_blocks_parities = [
            (
                block_size,
                int(np.ceil(message_size / block_size)),
            )
            for block_size in block_sizes
        ]

        # Remaining passes with random permutations
        block_size = min(
            message_size,
            int(np.ceil(message_size * self.sampling_fraction)),
        )
        size_blocks_parities.extend(
            [
                (
                    block_size,
                    int(np.ceil(message_size / block_size)),
                )
            ]
            * (self.number_of_passes - self.switch_after_pass)
        )
        return tuple(size_blocks_parities)



"""Test the Cascade error correction protocol."""

from __future__ import annotations

import pytest

from tno.quantum.communication.qkd_key_rate.classical import (
    Message,
    ParityStrategy,
    Permutations,
)
from tno.quantum.communication.qkd_key_rate.classical.cascade import (
    CascadeCorrector,
    CascadeReceiver,
    CascadeSender,
)
from tno.quantum.utils.validation import check_random_state

random_state = check_random_state(None, "random_state")


@pytest.mark.parametrize("switch_after_pass", [None, 3])
def test_correctness(switch_after_pass: int | None) -> None:
    """
    Test correctness for two different ParityStrategies
        - Apply only normal passes
        - Switching after 3 passes
    """

    message_length = 100000
    error_rate = 0.05
    input_message = Message(
        [int(random_state.rand() > 1 / 2) for _ in range(message_length)]
    )
    error_message = input_message.apply_errors(error_rate=error_rate)

    number_of_passes = 8
    sampling_fraction = 0.34
    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_length
    )
    parity_strategy = ParityStrategy(
        error_rate=error_rate,
        sampling_fraction=sampling_fraction,
        number_of_passes=number_of_passes,
        switch_after_pass=switch_after_pass,
    )

    alice = CascadeSender(message=input_message, permutations=permutations)
    bob = CascadeReceiver(
        message=error_message,
        permutations=permutations,
        parity_strategy=parity_strategy,
    )

    corrector = CascadeCorrector(alice=alice, bob=bob)
    results = corrector.correct_errors()

    assert results.output_alice == results.output_bob
    assert results.output_alice.length == message_length
    assert alice.number_of_exposed_bits == bob.number_of_exposed_bits



r"""Privacy amplification module.

The :py:class:`~tno.quantum.communication.qkd_key_rate.classical.privacy_amplification.PrivacyAmplification`
class can be used to compute the hash of an error corrected string. The length of the
hashed string equals the remaining entropy.

Typical usage example:

    >>> from tno.quantum.communication.qkd_key_rate.classical.privacy_amplification import (
    ...    PrivacyAmplification,
    ... )
    >>>
    >>> message = Message.random_message(message_length=100)
    >>> privacy = PrivacyAmplification(message.length, error_rate_basis_x=0)
    >>> entropy = privacy.get_entropy_estimate(error_correction_loss=10)
    >>> privacy.do_hash(message, entropy)  # doctest: +SKIP
    b'\x11\x8f\xba\xc8\xc2\x97\xce\xd7\xf0\xf4F\xd1\xfew\xf7\xd0\xd2Y>\x08J\x1e\x1a\xb9\x1d.\x8d8\xc5\x01\xa7(u\xc0\xcd\x1bc\xfck\xbc_!.\xea\xf5v\xd9\x90\xd4a\x89,\xdaZ\xf6tq\xdf\xe77\x16\x07\x15\xdc\x8d\x86`\x80Wy\x7fHU\xc8\xe3\xe7\xcf \xe6V\x8f\x19\x1c!sv\xad\x9b\xc4\x1c'

"""  # noqa: E501

import hashlib

from tno.quantum.communication.qkd_key_rate._utils import (
    one_minus_binary_entropy as one_minus_h,
)
from tno.quantum.communication.qkd_key_rate.classical._message import Message


class PrivacyAmplification:
    """Privacy amplification.

    A hash of an error corrected string is computed and returned. The length of the
    hashed string equals the remaining entropy.
    """

    def __init__(self, observed_pulses_basis_x: int, error_rate_basis_x: float) -> None:
        """Init of PrivacyAmplification.

        Args:
            observed_pulses_basis_x: Number of pulses received in the X-basis
            error_rate_basis_x: Error rate in the pulses received in the X-basis
        """
        self.observed_pulses_basis_x = observed_pulses_basis_x
        self.error_rate_basis_x = error_rate_basis_x

    def get_entropy_estimate(self, error_correction_loss: int = 0) -> float:
        """Estimate the amount of entropy.

        Uses the key-rate estimation functions to determine the key-rate and
        obtains the number of secure bits by multiplying this by the number of
        pulses sent. Adjusts the remaining entropy for losses due to the
        error-correction.

        Args:
            error_correction_loss: Number of corrected errors

        Returns:
            The amount of entropy
        """
        entropy = self.observed_pulses_basis_x * float(
            one_minus_h(self.error_rate_basis_x)
        )
        entropy -= error_correction_loss

        return entropy

    def do_hash(self, message: Message, entropy: float) -> bytes:
        """Computes the hash given a message of bits.

        The length of the hash equals the secure entropy we have.
        """
        if entropy < 0:
            error_msg = "Entropy is smaller than 0. Secure hash cannot be computed"
            raise ValueError(error_msg)

        hash_function = hashlib.shake_256()
        hash_function.update(bytes(message))

        return hash_function.digest(int(entropy))



"""Test the Message object."""

from copy import deepcopy

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical import Message


def test_random_message_init() -> None:
    """Test creating random permutation"""
    message_size = 1000
    message = Message.random_message(message_length=message_size)

    assert isinstance(message, Message)
    assert message.length == message_size


def test_getter() -> None:
    """Test getter"""
    original_message = [1, 1, 0, 0, 1, 0, 1, 0, 1]
    message = Message(message=original_message)

    for i in range(message.length):
        assert original_message[i] == message[i]


def test_bytes() -> None:
    """Test bytes"""
    message = Message(message=[1, 1, 0, 1])
    assert bytes(message) == b"1101"


def test_permutation() -> None:
    """Test apply permutation"""
    raw_message = [1, 1, 1, 1, 1, 0, 0, 0, 0, 0]
    permutation = [5, 6, 7, 8, 9, 0, 1, 2, 3, 4]

    message = Message(message=raw_message)
    message.apply_permutation(permutation)

    assert message.message == [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]

    message1 = Message.random_message(message_length=10)
    message2 = deepcopy(message1)
    inv_permutation: list[int] = np.argsort(permutation).tolist()
    message2.apply_permutation(permutation)
    message2.apply_permutation(inv_permutation)
    assert message1 == message2


def test_apply_error() -> None:
    """Test apply error."""
    message_length = 100_000
    message = Message.random_message(message_length=message_length)
    assert message.apply_errors(error_rate=1.0) == Message([not x for x in message])
    assert message.apply_errors(error_rate=0.0) == message

    error_rate = 0.5
    error_message = message.apply_errors(error_rate)
    number_of_errors = sum(x != y for (x, y) in zip(message, error_message))
    expected_number_of_errors = message_length * error_rate
    assert number_of_errors <= expected_number_of_errors + 0.1 * message_length
    assert expected_number_of_errors - 0.1 * message_length <= number_of_errors



"""Quantum Key Distribution (QKD) Protocols.

The following quantum protocols are supported:

- BB84 protocol (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bb84`),
- BB84 protocol using a single photon source (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bb84_single_photon`),
- BBM92 protocol (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bbm92`).

Each protocol requires the user to define the properties of the :py:class:`~tno.quantum.communication.qkd_key_rate.quantum.Detector`
that is being used by Bob. The BBM92 additionally allows information on the detector
used by Alice. By default we consider the following standard detector

.. code-block:: python

    from tno.quantum.communication.qkd_key_rate.quantum import Detector
    standard_detector = Detector(
        name="standard",
        efficiency_detector=0.2,
        jitter_source=0,
        jitter_detector=5.00e-11,
        dead_time=4.50e-08,
        dark_count_frequency=100,
        detection_frequency=1.00e07,
        detection_window=5,
        efficiency_system=1,
    )

Throughout the codebase the following units are used:

.. list-table::
   :header-rows: 1

   * - Variable Name
     - Unit
   * - name
     - Unitless (string)
   * - jitter_source
     - Seconds (s)
   * - jitter_detector
     - Seconds (s)
   * - dead_time
     - Seconds (s)
   * - detection_window
     - Seconds (s)
   * - efficiency_system
     - Unitless (percentage or fraction)
   * - polarization_drift
     - Radians (rad)
   * - error_detector
     - Unitless (percentage or fraction)
   * - dark_count_frequency
     - Counts per second (Hz)
   * - dark_count_rate
     - Counts per detection window (unitless)
   * - detection_frequency
     - Detection windows per second (Hz)
   * - interval
     - Seconds (s)
   * - efficiency_detector
     - Unitless (percentage or fraction)
   * - efficiency_party
     - Unitless (percentage or fraction)
   * - Key-rate
     - Bits per second (bps)
   * - mu (Intensities of the laser)
     - Unitless (mean photon number per pulse)
   * - attenuation (Loss of the channel)
     - Decibels (dB)


Usage examples of the various protocols can be found in the relevant submodules.
"""  # noqa: E501

from tno.quantum.communication.qkd_key_rate.quantum._detector import (
    Detector,
    standard_detector,
)
from tno.quantum.communication.qkd_key_rate.quantum._keyrate import KeyRate

__all__ = [
    "Detector",
    "KeyRate",
    "standard_detector",
]



"""Util functions for calculating the efficiency and communication rounds."""

from __future__ import annotations

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical._parity_strategy import (
    ERROR_RATES_TO_BLOCK_SIZES,
)
from tno.quantum.communication.qkd_key_rate.classical._schedule import (
    ERROR_RATES_TO_SCHEDULE,
)

EFFICIENCY = {
    "winnow": {
        0.010: 0.90,
        0.020: 0.83,
        0.030: 0.80,
        0.040: 0.71,
        0.050: 0.65,
        0.060: 0.59,
        0.070: 0.54,
        0.080: 0.49,
        0.090: 0.44,
        0.100: 0.40,
        0.110: 0.34,
        0.120: 0.29,
        0.130: 0.27,
        0.140: 0.22,
        0.150: 0.19,
    },
    "cascade": {
        0.010: 0.83,
        0.020: 0.78,
        0.030: 0.73,
        0.040: 0.67,
        0.050: 0.63,
        0.060: 0.58,
        0.070: 0.55,
        0.080: 0.51,
        0.090: 0.47,
        0.100: 0.44,
        0.110: 0.39,
        0.120: 0.35,
        0.130: 0.32,
        0.140: 0.30,
        0.150: 0.27,
    },
}


def compute_efficiency(method: str, error_rate: float) -> float:
    """Calculate the efficiency of the error correction protocol.

    If the error-rate is above maximum defined error rate, set the efficiency to 0.20.

    Args:
        method: the error correction protocol (``'Winnow'`` or ``'Cascade'``)
        error_rate: estimated error rate

    Returns:
        Efficiency of the error correction protocol

    Raises:
        ValueError: If unknown method is provided.
    """
    try:
        efficiency_dict = EFFICIENCY[method.lower()]
    except KeyError as exc:
        error_msg = "Unknown passed corrector-method."
        raise ValueError(error_msg) from exc

    error_keys = efficiency_dict.keys()
    try:
        er = min(x for x in error_keys if x >= error_rate)
    except ValueError:
        return 0.20
    return efficiency_dict.get(er, 0.20)


def compute_estimate_on_communication(
    method: str, error_rate: float, message_length: int | None = None
) -> int:
    """Estimate on the number of communication rounds.

    Args:
        method: the error correction protocol (``'Winnow'`` or ``'Cascade'``)
        error_rate: the error rate
        message_length: Length of the message, only required for cascade protocol

    Returns:
        Estimate for the number of communication rounds. For the Cascade protocol only a
        rough upper bound can be provided.

    Raises:
        ValueError: If error rate is to high for a secure protocol.
        ValueError: If message length is not provided with Cascade protocol.
        ValueError: If provided method is not among valid options.
    """
    if method.lower() == "winnow":
        error_schedule_keys = ERROR_RATES_TO_SCHEDULE.keys()
        try:
            er = min(x for x in error_schedule_keys if x >= error_rate)
        except ValueError as exc:
            error_msg = "Error rate too high for secure protocol"
            raise ValueError(error_msg) from exc

        return 1 + 2 * sum(ERROR_RATES_TO_SCHEDULE[er])

    if method.lower() == "cascade":
        # Only a loose estimate on the number of communications can be given.
        if message_length is None:
            error_msg = "Message length needs to be provided for Cascade protocol"
            raise ValueError(error_msg)

        error_block_keys = ERROR_RATES_TO_BLOCK_SIZES.keys()
        try:
            er = min(x for x in error_block_keys if x >= error_rate)
        except ValueError as exc:
            error_msg = "Error rate too high for secure protocol"
            raise ValueError(error_msg) from exc

        start_block_size = ERROR_RATES_TO_BLOCK_SIZES[er]
        number_of_passes = 8
        block_sizes = [start_block_size * 2**i for i in range(number_of_passes)]
        max_sizes = np.repeat([message_length // 2], number_of_passes)
        block_sizes = np.vstack([block_sizes, max_sizes]).min(axis=0)

        return int(
            1
            + 2
            * np.sum(
                [
                    np.sum(np.ceil(np.log2(block_sizes[: i + 1])))
                    for i in range(number_of_passes)
                ],
                dtype=int,
            )
        )

    error_msg = "Unknown passed corrector-method."
    raise ValueError(error_msg)



"""Test Corrector base functions"""

import hashlib
import hmac
from copy import deepcopy

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical import (
    Corrector,
    Message,
    ParityStrategy,
    Permutations,
)
from tno.quantum.communication.qkd_key_rate.classical._schedule import (
    Schedule,
)
from tno.quantum.communication.qkd_key_rate.classical.cascade import (
    CascadeCorrector,
    CascadeReceiver,
    CascadeSender,
)
from tno.quantum.communication.qkd_key_rate.classical.winnow import (
    WinnowCorrector,
    WinnowReceiver,
    WinnowSender,
)
from tno.quantum.utils.validation import check_random_state

random_state = check_random_state(None, "random_state")


def test_calculate_number_of_errors() -> None:
    """Test calculation of number of errors"""
    message1 = Message.random_message(100)
    message2 = deepcopy(message1)
    assert Corrector.calculate_number_of_errors(message1, message2) == 0
    for i in range(message1.length):
        message2[i] = 1 - message1[i]
        assert Corrector.calculate_number_of_errors(message1, message2) == i + 1


def test_calculate_error_rate() -> None:
    """Test calculation of ratio number of errors"""
    message_length = 100
    message1 = Message.random_message(message_length)
    message2 = deepcopy(message1)
    assert Corrector.calculate_error_rate(message1, message2) == 0
    for i in range(message1.length):
        message2[i] = 1 - message1[i]
        assert (
            Corrector.calculate_error_rate(message1, message2)
            == (i + 1) / message_length
        )


def test_calculate_key_reconciliation_rate_cascade() -> None:
    """
    Test key reconciliation rate calculation with Cascade.
    """
    message_length = 1000
    error_rate = 0.01
    input_message = Message(
        [int(random_state.rand() > 1 / 2) for _ in range(message_length)]
    )
    error_message = input_message.apply_errors(error_rate=error_rate)

    number_of_passes = 5
    sampling_fraction = 0.5
    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_length
    )
    parity_strategy = ParityStrategy(
        error_rate=error_rate,
        sampling_fraction=sampling_fraction,
        number_of_passes=number_of_passes,
        switch_after_pass=number_of_passes,
    )

    alice = CascadeSender(message=input_message, permutations=permutations)
    bob = CascadeReceiver(
        message=error_message,
        permutations=permutations,
        parity_strategy=parity_strategy,
    )

    corrector = CascadeCorrector(alice=alice, bob=bob)
    results = corrector.correct_errors()

    assert (
        results.key_reconciliation_rate
        == corrector.calculate_key_reconciliation_rate(exposed_bits=True)
    )


def test_calculate_key_reconciliation_rate_winnow() -> None:
    """
    Test key reconciliation rate calculation with Winnow.
    """
    message_length = 1000
    error_rate = 0.01
    input_message = Message.random_message(message_length=message_length)
    error_message = input_message.apply_errors(error_rate=error_rate)

    schedule = Schedule.schedule_from_error_rate(error_rate=error_rate)
    number_of_passes = np.sum(schedule.schedule)
    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_length
    )

    alice = WinnowSender(
        message=input_message, permutations=permutations, schedule=schedule
    )
    bob = WinnowReceiver(
        message=error_message, permutations=permutations, schedule=schedule
    )

    corrector = WinnowCorrector(alice=alice, bob=bob)
    results = corrector.correct_errors()

    assert (
        results.key_reconciliation_rate
        == corrector.calculate_key_reconciliation_rate(exposed_bits=True)
    )


def test_calculate_key_reconciliation_rate_artificial() -> None:
    """
    Test key reconciliation rate calculation without running a protocol.
    """
    message_length = 1000
    exposed_bits = 150
    error_rate = 0.001
    input_message = Message.random_message(message_length=message_length)
    error_message = deepcopy(input_message)

    number_of_passes = 5
    sampling_fraction = 0.5
    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_length
    )
    parity_strategy = ParityStrategy(
        error_rate=error_rate,
        sampling_fraction=sampling_fraction,
        number_of_passes=number_of_passes,
        switch_after_pass=number_of_passes,
    )

    alice = CascadeSender(message=input_message, permutations=permutations)
    bob = CascadeReceiver(
        message=error_message,
        permutations=permutations,
        parity_strategy=parity_strategy,
    )

    corrector = CascadeCorrector(alice=alice, bob=bob)
    corrector.alice._number_of_exposed_bits = exposed_bits
    corrector.alice._net_exposed_bits = exposed_bits

    assert corrector.calculate_key_reconciliation_rate(exposed_bits=True) == (
        1 - exposed_bits / message_length
    )


def test_create_message_tag_pair() -> None:
    """Test message-tag hashed pair"""
    message_length = 1000
    input_message = Message(
        [int(random_state.rand() > 1 / 2) for _ in range(message_length)]
    )

    shared_key = "key"
    message, tag = CascadeCorrector.create_message_tag_pair(
        input_message, shared_key=shared_key
    )

    message_string = "".join(str(x) for x in input_message.message)
    assert message == bytes(message_string.encode("utf-8"))
    assert (
        tag
        == hmac.new(
            key=bytes(shared_key.encode("utf-8")), msg=message, digestmod=hashlib.sha384
        ).digest()
    )



"""Classical Error-Correction Modules.

The following classical error-correction protocols are supported:

- Cascade (:py:mod:`~tno.quantum.communication.qkd_key_rate.classical.cascade`),
- Winnow (:py:mod:`~tno.quantum.communication.qkd_key_rate.classical.winnow`).

The presented code can be used to:

- Correct errors in exchanged sifted keys for different QKD protocols,
- Enhance the reliability and security of the generated keys.

Usage examples can be found in the relevant submodules.
"""

# isort: skip_file

from tno.quantum.communication.qkd_key_rate.classical._message import (
    Message,
)
from tno.quantum.communication.qkd_key_rate.classical._permutations import (
    Permutations,
)
from tno.quantum.communication.qkd_key_rate.classical._schedule import (
    Schedule,
)
from tno.quantum.communication.qkd_key_rate.classical._parity_strategy import (
    ParityStrategy,
)

from tno.quantum.communication.qkd_key_rate.classical._sender import (
    SenderBase,
)
from tno.quantum.communication.qkd_key_rate.classical._receiver import (
    ReceiverBase,
)
from tno.quantum.communication.qkd_key_rate.classical._corrector import (
    Corrector,
    CorrectorOutputBase,
)
from tno.quantum.communication.qkd_key_rate.classical._required_linktime import (
    compute_estimate_on_communication,
    compute_efficiency,
)


__all__ = [
    "Corrector",
    "CorrectorOutputBase",
    "Message",
    "ParityStrategy",
    "Permutations",
    "ReceiverBase",
    "Schedule",
    "SenderBase",
    "compute_efficiency",
    "compute_estimate_on_communication",
]



"""Classes to perform key error rate estimate for the BBM92 QKD protocol.

Quantum Key Distribution (QKD) protocols rely on an entangled photon source that
produces entangled photon pairs, which are distributed over two parties. Both parties
randomly choose one of two predefined measurement bases. As the photons are entangled,
non-random results will only be obtained for specific combinations of basis choices.
Detection events are sifted, where only the detection events corresponding to events
where both parties measured in an appropriate basis should be kept. Part of the
resulting sifted key is made public which can be used to estimate the key error rate.

The BBM92 QKD protocol can be seen as a generalization of the famous BB84 protocol.
The protocols differ from each other by using entangled photon pairs that are sent
to two parties. Due to the entanglement, the measurements give both parties information
of each others outcome and allows them to create the key. The exact relation between
the measurement results depends on how the entanglement is created. We have an
tunable source that lets you modify the expected number of entangled photon pairs.
If one party was located at the source and directly measures its quantum state, we
effectively have the BB84 protocol again.


We consider two cases

- Asymptotic Key rate:
    The number of pulses is asymptotic and we only have to optimize the intensity of
    the source. As we have no decoy-states, we do not have to optimize probabilities
    for a measurement basis. For the gain and error-rate computation we include the
    link effects for both parties.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bbm92 import (
    ...     BBM92AsymptoticKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> finite_key_rate = BBM92AsymptoticKeyRateEstimate(detector=detector)
    >>> x, rate = finite_key_rate.optimize_rate(attenuation=0.2)
    >>> print(f"{x['mu']=}, {rate=}")  # doctest: +SKIP
    x['mu']=array([0.04789677]), rate=0.0013733145512722993

- Finite Key rate:
    The number of pulses is finite and chosen by the user. We optimize both the
    intensity of the source and the probability of measuring in the X and Z basis.
    We take finite-key effects into account by computing bounds on the expected number
    of states in a basis and we use security parameters (set as default, but changeable
    by the user) to tune the confidence of these bounds.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bbm92 import (
    ...     BBM92FiniteKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> finite_key_rate = BBM92FiniteKeyRateEstimate(detector=detector)
    >>> x, rate = finite_key_rate.optimize_rate(attenuation=0.2)
    >>> print(f"{x['mu']=}, {rate=}")  # doctest: +SKIP
    x['mu']=array([0.04781671]), rate=0.0007179260057907571
"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

import numpy as np
import scipy
from numpy.typing import ArrayLike, NDArray

from tno.quantum.communication.qkd_key_rate._utils import binary_entropy as h
from tno.quantum.communication.qkd_key_rate._utils import (
    one_minus_binary_entropy as one_minus_h,
)
from tno.quantum.communication.qkd_key_rate.quantum._config import (
    NLP_CONFIG,
    OptimizationError,
)
from tno.quantum.communication.qkd_key_rate.quantum._keyrate import (
    AsymptoticKeyRateEstimate,
    FiniteKeyRateEstimate,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.quantum import Detector

warnings.filterwarnings(
    "ignore", message="delta_grad == 0.0. Check if the approximated function is linear."
)


def efficiency_channel(attenuation: float) -> float:
    """Calculates the efficiency of the channel.

    Args:
        attenuation: Loss in dB for the channel

    Returns:
        Efficiency channel
    """
    return float(np.power(10, -attenuation / 10))


def efficiency_system(efficiency: float, attenuation: float) -> float:
    """Calculates the efficiency of the system.

    This includes the efficiency of the communication link.

    Args:
        efficiency: Efficiency of source and detection apparatus
        attenuation: Loss in dB for the channel

    Returns:
        Efficiency system
    """
    return efficiency * efficiency_channel(attenuation)


def compute_gain_and_error_rate(
    detector_alice: Detector,
    detector_bob: Detector,
    attenuation_alice: float,
    attenuation_bob: float,
    half_mu: float,
) -> tuple[float, float]:
    r"""Computes the total gain and error rate of the channel.

    Args:
        detector_alice: Alice's detector
        detector_bob: Bob's detector
        attenuation_alice: Attenuation of Alice's channel
        attenuation_bob: Attenuation of Bob's channel
        half_mu: Half of the expected number of photon-pairs

    Returns:
        - The overall gain. The probability of a coincident event, given a pump pulse.
        - The overall error-rate.

    Errors can result from incorrect states generated, losses over the channel and
    losses at the detector side.

    - error_background:

        Error rate related to background errors, with error-rate 1/2

    - error_detector:

        Intrinsic detector errors. The probability that a photon hits the erroneous
        detector, this error indicates the relative error between the detectors at both
        sides.

    The initial states are generated using parametric down conversion (PDC), which gives
    the probability

    .. math::
        P(n) = (n+1) \cdot \lambda^n / [(1+\lambda)^{n+2}],

    to generate an $n$-photon-pair. The variable $\lambda$ corresponds to half the
    expected photon pair number.

    To compute the losses over the channel, we compute the overall transmittance of the
    channel, which is given by

    .. math::
        \eta_n = [1 - (1-\eta_A)^n] \cdot [1 - (1-\eta_B)^n],

    where:

        - $\eta_n$: The overall transmittance,
        - $\eta_x$: The channel loss of $x$.

    The probability of a detection event, conditional on an $n$-photon-pair being
    emitted is given by the yield:

    .. math::
        {Yield}_n = [1 - (1-Y_{0A}) \cdot (1-\eta_{A})^n] \cdot
        [1 - (1-Y_{0B}) \cdot (1-\eta_{B})^n],

    where:

        - $Y_{0x}$: The darkcount rate of $x$.

    The overall contribution of $n$-photon-pairs to the number of detected events is
    given by the product of the yield and probability:

    .. math::
        Q_n = Y_n \cdot P(n),

    where:

        - $Q_n$: The gain of the n-photon-pair.

    and the overall gain is then given by the sum of these individual gains. The overall
    gain corresponds to the probability of a detection event, conditional on a pulse
    being sent

    .. math::
        Q_\lambda = \sum_{n=0}^{\infty} Q_n.

    From this we compute the quantum bit-error rate (QBER):

    .. math::
        E_\lambda \cdot Q_\lambda = \sum_{n=0}^{\infty} e_n \cdot Y_n \cdot P(n),

    where

        - $e_n$: The error-rate for n-photon-pair states.

    The functions below compute $Q_{\lambda}$ and $E_{\lambda}$ and return them. We
    optimized the implementation of the functions and by that deviated from standard
    implementations in literature. Our implementation is equivalent, but gives better
    precision in high attenuation cases.

    Formulas derived from "Quantum key distribution with entangled photon sources"
    (doi: `10.1103/PhysRevA.76.012307`_)

    .. _10.1103/PhysRevA.76.012307: http://doi.org/10.1103/PhysRevA.76.012307
    """
    dark_count_rate_alice = detector_alice.dark_count_rate
    dark_count_rate_bob = detector_bob.dark_count_rate
    channel_efficiency_alice = efficiency_system(
        detector_alice.efficiency_party, attenuation_alice
    )
    channel_efficiency_bob = efficiency_system(
        detector_bob.efficiency_party, attenuation_bob
    )

    # Errors from background noise are uniform, thus equal to 0.5
    # Errors from the detector depend on the detector
    error_background = 0.5
    if hasattr(detector_alice, "error_detector"):
        error_detector = detector_alice.error_detector
    elif hasattr(detector_bob, "error_detector"):
        error_detector = detector_bob.error_detector
    else:
        error_detector = 0

    a = dark_count_rate_alice
    b = dark_count_rate_bob
    A = channel_efficiency_alice * half_mu
    B = channel_efficiency_bob * half_mu
    C = channel_efficiency_alice * channel_efficiency_bob * half_mu
    D = A + B - C
    # Note that this implementation, though equivalent, differs from literature
    # This however gives better results for high attenuations and hence for
    # small A, B and C
    error_rate_denominator = (
        (2 * A + A**2 + a) * (1 + B) * (1 + D) / (1 + A)
        + (2 * B + B**2 + b) * (1 + A) * (1 + D) / (1 + B)
        + (a * b - a - b - 2 * D - D**2) * (1 + A) * (1 + B) / (1 + D)
    )

    error_rate_numerator = (
        2
        * (error_background - error_detector)
        * channel_efficiency_alice
        * channel_efficiency_bob
        * half_mu
        * (1 + half_mu)
    )

    # Computing the overall gain and error rate
    gain_lambda = (
        (2 * A + A**2 + a) / np.power(1 + A, 2)
        + (2 * B + B**2 + b) / np.power(1 + B, 2)
        + (a * b - a - b - 2 * D - D**2) / np.power(1 + D, 2)
    )

    error_rate_lambda = error_background - error_rate_numerator / error_rate_denominator

    return gain_lambda, error_rate_lambda


def delta(N: float, delta_bit: float, e1: float) -> float:
    """Gives the upper bound delta in the finite case.

    Args:
        N: Number of detection events
        delta_bit: Error in the X-basis
        e1: Epsilon-security parameter

    Returns:
        Delta
    """
    return float(
        2 * np.sqrt(np.max([-delta_bit * (1 - delta_bit) * np.log(e1) / N, 0]))
    )


class BBM92AsymptoticKeyRateEstimate(AsymptoticKeyRateEstimate):
    """BBM92 Asymptotic Key Rate."""

    def __init__(
        self,
        detector: Detector,
        detector_alice: Detector | None = None,
        **kwargs: Any,
    ) -> None:
        """Class for asymptotic key-rate estimation.

        Args:
            detector: The detector used at Bob's side
            detector_alice: The detector used at Alice's side.
                default same detector as Bob.
            kwargs: protocol specific arguments.
        """
        super().__init__(detector=detector, args=kwargs)
        self.detector_alice = detector if detector_alice is None else detector_alice

        self.last_positive: float = -1
        self.last_x: NDArray[np.float64]

    def compute_rate(self, mu: float | ArrayLike, attenuation: float) -> float:  # type: ignore[override]
        """Compute the key-rate for a specific intensity and attenuation.

        Note, the source can be placed anywhere between the two parties. If the
        source is placed at one party, we effectively have the BB84 protocol.
        The optimal gain is achieved with the source placed in middle between
        the two parties.

        Args:
            mu: Intensity
            attenuation: Attenuation

        Returns:
            Key-rate

        Raises:
            ValueError: When mu is given with invalid dimensions
        """
        scale_factor = float(np.power(10, attenuation))

        mu = np.atleast_1d(mu)
        if mu.shape != (1,):
            error_msg = "Multiple intensities were given, only one is expected."
            raise ValueError(error_msg)

        gain_lambda, error_rate_lambda = compute_gain_and_error_rate(
            self.detector,
            self.detector_alice,
            attenuation / 2,
            attenuation / 2,
            float(mu) / 2,
        )

        # These are the error rates corresponding to the X and Z basis
        # For the asymptotic case, the two are the same
        delta_bit = error_rate_lambda
        delta_phase = error_rate_lambda

        key_rate = float(
            scale_factor * gain_lambda * (one_minus_h(delta_phase) - h(delta_bit))
        )

        self.last_positive = key_rate
        self.last_x = np.hstack([mu])
        return key_rate / scale_factor

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        return {"mu": x}

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value, default value [0.5] is chosen.
            bounds: Bounds on search range, default [(0.001, 0.5)]

        Returns:
            Optimized intensity and key-rate

        Raises:
            ValueError: When x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
        """
        if bounds is None:
            # The lower and upper bound on the laser intensity
            lower_bound = np.array([0.001])
            upper_bound = np.array([0.9])
        else:
            if len(bounds) != 1:
                error_msg = (
                    "Invalid dimensions input bounds. Expected 1 upper- and lower"
                    f" bound but received {len(bounds)} bounds."
                )
                raise ValueError(error_msg)
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
        bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        if x0 is None:
            x0 = [0.5]
        elif len(np.asarray(x0)) != 1:
            error_msg = (
                "Invalid number of inputs. Expected 1 intensity."
                f" Received {len(np.asarray(x0))} intensities."
            )
            raise ValueError(error_msg)

        self.last_positive = -1
        self.last_x = np.asarray(x0)

        args = {"attenuation": attenuation}
        res = scipy.optimize.minimize(
            self._f, x0, args=args, bounds=bounds, **NLP_CONFIG
        )

        mu = res.x
        rate = -res.fun
        if rate < 0:
            error_msg = "Optimization resulted in a negative key rate."
            raise ValueError(error_msg)
        return self._extract_parameters(mu), float(rate)


class BBM92FiniteKeyRateEstimate(FiniteKeyRateEstimate):
    """BBM92 Finite key-rate."""

    def __init__(
        self,
        detector: Detector,
        number_of_pulses: int = int(1e12),
        detector_alice: Detector | None = None,
        **kwargs: Any,
    ) -> None:
        """Class for finite key-rate estimation.

        Args:
            detector: The detector used at Bob's side
            number_of_pulses: Number of pulses sent in total
            detector_alice: The detector used at Alice's side.
                default same detector as Bob.
            kwargs: Protocol specific input.
        """
        super().__init__(detector=detector, args=kwargs)
        self.number_of_pulses = number_of_pulses

        self.last_positive: float = -1
        self.last_x: NDArray[np.float64]

        self.detector_alice = detector if detector_alice is None else detector_alice

    def _compute_last_positive_distance(self, x: NDArray[np.float64]) -> float:
        """Computes the last positive distance.

        The optimization routine sometimes considers a parameter setting
        outside of the valid region. This function is used to push the
        parameters back to the valid regime.
        """
        if self.last_positive > -1:
            return self.last_positive - float(np.linalg.norm(x - self.last_x))
        return self.last_positive

    def compute_rate(  # type: ignore[override]
        self,
        mu: float | ArrayLike,
        attenuation: float,
        probability_basis_X: ArrayLike,
        probability_basis_Z: ArrayLike,
        n_X: int | None = None,
    ) -> float:
        """Compute the key-rate for a specific set of variables.

        Note, the source can be placed anywhere between the two parties. If the
        source is placed at one party, we effectively have the BB84 protocol.
        The optimal gain is achieved with the source placed in middle between
        the two parties.

        Args:
            mu: Used intensities
            attenuation: Attenuation of the channel
            probability_basis_X: Probabilities for each intensity in X-basis
            probability_basis_Z: Probabilities for each intensity in Z-basis
            n_X: Number of pulses in the X-basis

        Returns:
            key-rate

        Raises:
            ValueError: When mu is given with invalid dimensions.
        """
        probability_basis_X = np.asarray(probability_basis_X)
        probability_basis_Z = np.asarray(probability_basis_Z)
        scale_factor = np.power(10, attenuation / 10)

        mu = np.atleast_1d(mu)
        if mu.shape != (1,):
            error_msg = "Multiple intensities were given, only one is expected."
            raise ValueError(error_msg)

        # Two security parameters to be used
        p_abort = 2**-50
        epsilon_1 = 1e-12

        # Make sure the variables are within a valid range
        x = np.hstack((mu, probability_basis_X, probability_basis_Z))
        if (
            x.min() < 0
            or x.max() > 1
            or (probability_basis_X + probability_basis_Z) != 1
        ):
            return self._compute_last_positive_distance(x)

        gain_lambda, error_rate_lambda = compute_gain_and_error_rate(
            self.detector,
            self.detector_alice,
            attenuation / 2,
            attenuation / 2,
            float(mu / 2),
        )

        # Compute the number of pulses in the X basis. The Z-basis follows
        n_X_observed = (
            gain_lambda * probability_basis_X * self.number_of_pulses
            if n_X is None
            else n_X
        )

        # Check if the number of pulses in the X basis is positive
        if n_X_observed <= 0:
            return self._compute_last_positive_distance(x)

        # These are the error rates corresponding to the X and Z basis
        # A bound is used for delta_phase (Z) errors
        delta_bit = error_rate_lambda
        delta_phase = error_rate_lambda + delta(
            gain_lambda * self.number_of_pulses, delta_bit, epsilon_1
        )

        # Only the pulses in the X-basis are used to get key-material
        usable_pulses_lp = n_X_observed * (one_minus_h(delta_phase) - h(delta_bit))
        key_rate = float(
            scale_factor * (1 - p_abort) * usable_pulses_lp / self.number_of_pulses
        )

        self.last_positive = key_rate
        self.last_x = x
        return self.last_positive

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        """Extract the parameters and assigns them correspondingly."""
        return {
            "mu": np.array([x[0]]),
            "probability_basis_X": np.array([x[1]]),
            "probability_basis_Z": np.array([x[2]]),
        }

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value
            bounds: Bounds on search range
        Returns:
            Optimized x=[intensity, probability_basis_X, probability_basis_Z]
            and found optimal key-rate

        Raises:
            ValueError: when x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
        """
        if bounds is None:
            lower_bound = np.array([0.01, 0, 0])
            upper_bound = np.array([0.8, 1, 1])
        else:
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
            if len(lower_bound) != 3 or len(upper_bound) != 3:
                error_msg = (
                    "Invalid dimensions input bounds. Expected 3 upper- and lower"
                    f" bound. Received {len(lower_bound)} lower- and {len(upper_bound)}"
                    f" upper bounds."
                )
                raise ValueError(error_msg)

        bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        if x0 is None:
            x0 = (lower_bound + upper_bound) / 2
        x0 = np.asarray(x0, dtype=float)
        if len(x0) != 3:
            error_msg = (
                "Invalid number of inputs. Expected 3 inputs."
                f"Received {len(x0)} inputs."
            )
            raise ValueError(error_msg)
        # The probabilities are normalized to 1
        x0[1:] /= x0[1:].sum()

        self.last_positive = -1.0
        self.last_x = x0

        bounds = scipy.optimize.Bounds(lower_bound, upper_bound)
        # This forces the probabilities to sum to one
        b = np.ones(1)
        A = np.array([0, 1, 1])
        B = np.ones(1)

        constraint = scipy.optimize.LinearConstraint(A, b, B)

        args = {"attenuation": attenuation}
        num_attempts = 3
        for _ in range(num_attempts):  # Maximum 3 retries
            try:
                res = scipy.optimize.minimize(
                    self._f,
                    x0,
                    args=args,
                    constraints=constraint,
                    bounds=bounds,
                    **NLP_CONFIG,
                )
            except OptimizationError:  # noqa: PERF203 Redo calculation
                args["attenuation"] += 1e-8
            else:
                rate = np.atleast_1d(-res.fun)[0]
                if rate < 0:  # Retry when negative key-rate is found.
                    args["attenuation"] += 1e-8
                    continue
                return self._extract_parameters(res.x), float(rate)
        error_msg = "Unable to find solution for optimal key rate."
        raise OptimizationError(error_msg)



"""Base class for Schedule object."""

from __future__ import annotations

ERROR_RATES_TO_SCHEDULE = {
    0.010: [1, 0, 1, 1, 1, 0, 0, 1],
    0.020: [1, 1, 2, 0, 0, 0, 0, 1],
    0.030: [1, 3, 0, 0, 0, 1, 0, 1],
    0.040: [2, 1, 1, 1, 0, 0, 0, 1],
    0.050: [2, 1, 1, 0, 0, 1, 1, 1],
    0.060: [2, 1, 2, 1, 0, 0, 0, 1],
    0.070: [2, 1, 2, 2, 1, 0, 0, 0],
    0.080: [2, 2, 1, 2, 0, 0, 0, 1],
    0.090: [2, 2, 1, 2, 1, 0, 0, 1],
    0.100: [3, 1, 2, 1, 0, 0, 0, 1],
    0.110: [3, 2, 1, 1, 0, 0, 0, 1],
    0.120: [3, 2, 3, 0, 0, 0, 0, 1],
    0.130: [4, 2, 2, 1, 0, 0, 0, 1],
    0.140: [4, 2, 2, 0, 0, 0, 0, 2],
    0.150: [4, 3, 2, 0, 0, 0, 1, 0],
}


class Schedule:
    """Schedule object for Winnow protocol."""

    def __init__(self, schedule: list[int]) -> None:
        """Init of Schedule.

        Args:
            schedule: a Winnow schedule.
        """
        self.schedule = schedule

        self.pass_number = 0
        self.processed_schedule = [0] * len(schedule)

    def __str__(self) -> str:
        """Schedule string representation."""
        return str(self.schedule)

    def __len__(self) -> int:
        """Length schedule."""
        return sum(self.schedule)

    def __eq__(self, other: object) -> bool:
        """Schedule equality."""
        if not isinstance(other, Schedule):
            return NotImplemented
        return self.schedule == other.schedule

    def next_pass(self) -> int:
        """Find iteration index for the next pass and update processed schedule.

        This iteration index describes the Hamming-distance used in that specific
        interaction in the protocol.

        Returns:
            iteration index
        """
        if self.pass_number >= len(self):
            error_msg = "Pass number is larger than length of schedule."
            raise IndexError(error_msg)
        self.pass_number += 1

        for i, schedule_i in enumerate(self.schedule):
            if self.processed_schedule[i] < schedule_i:
                self.processed_schedule[i] += 1
                return i
        return -1

    @property
    def remaining_passes(self) -> int:
        """Number of remaining passes in schedule."""
        return len(self) - sum(self.processed_schedule)

    @classmethod
    def schedule_from_error_rate(cls, error_rate: float = 0.15) -> Schedule:
        """Get schedule based on error rate."""
        error_keys = ERROR_RATES_TO_SCHEDULE.keys()
        try:
            min_error = min(x for x in error_keys if x >= error_rate)
        except ValueError as err:
            error_msg = "Error rate too high for secure protocol"
            raise ValueError(error_msg) from err

        schedule = ERROR_RATES_TO_SCHEDULE.get(min_error, [2, 1, 2, 2, 1, 0, 0, 0])
        return cls(schedule)



"""Base class for Receiving party."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from tno.quantum.communication.qkd_key_rate.classical import (
    Permutations,
    SenderBase,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.classical import Message


class ReceiverBase(SenderBase):
    """This class encodes all functions only available to the receiver.

    The receiver is assumed to have a string with errors and is thus assumed to
    correct the errors.
    """

    def __init__(
        self, message: Message, permutations: Permutations, name: str | None = None
    ) -> None:
        """Init of ReceiverBase.

        Args:
            message: Input message of the sender party
            permutations: List containing permutations for each pass
            name: Name of the sender party
        """
        super().__init__(message=message, permutations=permutations, name=name)

    @abc.abstractmethod
    def correct_errors(self, alice: SenderBase) -> None:
        """The main routine, find and correct errors."""

    def correct_individual_error(self, error_index: int) -> None:
        """Corrects a single error by flipping bit at location 'error_index'."""
        self._message[error_index] = 1 - self._message[error_index]



"""
Test entropy utils
"""

from tno.quantum.communication.qkd_key_rate._utils import binary_entropy
from tno.quantum.communication.qkd_key_rate._utils import (
    one_minus_binary_entropy as one_minus_h,
)


def test_one_minus_binary_entropy_zero_input() -> None:
    """Test one minus h for zero input"""
    assert one_minus_h(0) == 1


def test_one_minus_h() -> None:
    """Test one minus h on array"""
    assert all(one_minus_h([0.5, 0.5]) == [0.0, 0.0])


def test_binary_entropy_zero_input() -> None:
    """Test binary entropy function for zero input"""
    assert binary_entropy(0) == 0


def test_binary_entropy() -> None:
    """Test binary entropy function on array"""
    assert all(binary_entropy([0.5, 0.5]) == [1.0, 1.0])
